/*----------------------------------------------------------------------
PROGRAM: contrastmap.c
DATE:    6/28/94
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------


This code contains functions to contrast normalizes each image
according to the following:

   1. compute the (mean,std.dev.) of (possibly non-zero) pixels
   2. map the image into zero-mean unit-sigma 
   3. output image is this normalized image multiplied by a range
      factor and offset by an pedestal factor

---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <float.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

/*----------------------------------------------------------------------*/
/* Contrast_map just sends the image to contrast to be processed as */
 /* above, while checking that image values don't exceed 0 or 255 */


void contrast_map(srchparams_t sparms, image_t ret_image)
{

 
  register int i,j;
  int sets, bytes_pixel;
  int nrow, ncol;

 
  float scale = CONTRAST_SCALE;
  float pedestal = CONTRAST_PEDESTAL;
  float mean;
  float sigma;
  int npixels;

  /* ----  read indir descriptor file -------- */
  sets        = sparms.sets;
  bytes_pixel = sparms.bytes_pixel;
  ncol        = SCALEWARP_NCOL;
  nrow        = SCALEWARP_NROW;

  if (sets>1) 
    myerror("Input files must be single-set DAT files!");

  /* --- normalize the contrast --- */
      
  npixels = contrast(sparms.orig_image, ret_image, nrow, ncol, 
		     pedestal, scale, &mean, &sigma);

  /* --- Check output image --- */
  
  for (i=1; i<=nrow; i++) 
    for (j=1; j<=ncol; j++) 
      {
	if (ret_image[i][j]>255)
	  {
	    ret_image[i][j] = 255;
	  }
	else 
	  {
	    if (ret_image[i][j]<0)
	      ret_image[i][j] = 0;
	  }
      }
  return;
}


/* ====================================================================== */


int contrast(float **image_in, float **image_out, 
	      int nrow, int ncol,
	      float pedestal, float scale,
	      float *mean, float *sigma)
{
  register int i,j;
  float sum;
  float my_mean, my_sigma;
  int count;
  int nonzero_flag = NONZERO_FLAG;


  /* --- compute statistics ---- */

  sum = count = 0;
  for (i=1; i<=nrow; i++)
    for (j=1; j<=ncol; j++) {
      if (nonzero_flag == 1) {
	if (image_in[i][j] >= 1 ) {          /* This is > 1 rather than */
					    /* != 0 since files might */
					    /* be uchar or float */
	  sum += image_in[i][j];
	  count ++;
	}
      }
      else {
	sum += image_in[i][j];
	count ++;
      }
    }
  my_mean = sum/count;

  sum = count = 0;
  for (i=1; i<=nrow; i++)
    for (j=1; j<=ncol; j++) {
      if (nonzero_flag == 1) {
	if (image_in[i][j] > 1) {
	  sum += SQR(image_in[i][j] - my_mean);
	  count ++;
	}
      }
      else {
	sum += SQR(image_in[i][j] - my_mean);
	count ++;
      }
    }
  my_sigma = sqrt(sum/(count-1));


  /* --- remap input image ---- */

  for (i=1; i<=nrow; i++)
    for (j=1; j<=ncol; j++) 
      if (nonzero_flag == 1) {
	if (image_in[i][j] >= 1)
	  image_out[i][j] = pedestal + 
	    scale * (image_in[i][j]-my_mean)/my_sigma;
	else
	  image_out[i][j] = 0;
      }
  *mean = my_mean;
  *sigma = my_sigma;

  return count;
}







