/*----------------------------------------------------------------------
PROGRAM: eigsearch.c  
DATE:    6/16/94
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------
NOTE:    This version was adapted from ms_eigsearch.c
------------------------------------------------------------------------


  Local feature search by eigentemplates

  This routine uses data read in earlier from the following files:

  1. features.bf  which is an N-by-2 matrix defining the size (row,col)
     of each of the N templates.

  2. template[n].bf  where n=1...N is the eigenvector ROW matrix where
     the 1st row is the mean feature, and the remaining rows
     are the principal eigenvectors 1:M. 

     NOTE: The eigenvectors are stored in column-order (as in MATLAB)

  3. variances.bf which is an N-by-M matrix, where each row represents
     the rank-ordered eigenvalues associated with that feature

  4. windows.bf

  5. umap_musigma[n].bf

  6. spatial_musigma[n].bf


----------------------------------------------------------------------

   NOTE: In this version if unitmap'ing is selected it will use
         the Mahalanobis mean/stddev in addition to the standard
	 distance computation.

------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <float.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

extern void timestamp(char *label);

/* ---------------------------- MAIN ---------------------------------- */

imagparams_t eig_search(srchparams_t sparms, 
			int ***x_min, 
			int ***y_min,
			float ***err_min,
			int **n_min,
			float **variances,
			float **umap_musigma[MAX_NUM_TEMPLATES],
			float **spatial_musigma[MAX_NUM_TEMPLATES],
			float **w,
			int *num_minima)
     
{

  float ***template;
  float **errormap[MAX_NUM_TEMPLATES];
    
  int do_graymap = 0;  /*default is NOT to graymap each patch */
  int do_minima  = 0;  /* default is not to output minima files */
  int eig_first = 1;   /*search using first 5 eigenvectors by default */
  int eig_last  = 5;
  int singlefeature = 0;
  int do_unitmap;
  int do_mahalanobis;
  int do_correlation;
  int do_spatial;

  register int i,j,k,ii,jj;
  int c,feature=1,nframe,nfeatures,sets, bytes_pixel;
  int nrow, ncol, max_N;

  float fval1, fval, fval2, mean, sigma;

  int rowl[MAX_NUM_TEMPLATES],rowr[MAX_NUM_TEMPLATES];
  int coll[MAX_NUM_TEMPLATES],colr[MAX_NUM_TEMPLATES];
  int imin[MAX_NUM_TEMPLATES],imax[MAX_NUM_TEMPLATES];
  int jmin[MAX_NUM_TEMPLATES],jmax[MAX_NUM_TEMPLATES];
  int *N_dim; /* the dimensionality of each template */
  int *M_dim; /* the # of eigenvectors for each template */

  /* --- FIX: these should really be defined in pipe_main --- */
  float template_mean[MAX_NUM_TEMPLATES];   /* statistics for norm_corr */
  float template_sigma[MAX_NUM_TEMPLATES];  /* statistics for norm_corr */

  int rowm[MAX_NUM_TEMPLATES];
  int colm[MAX_NUM_TEMPLATES];
  float error[MAX_NUM_TEMPLATES];
  float *my_vector;
  
  int **templatesize;
  float *patch;

  struct Points myPoints_array[MAX_NUM_TEMPLATES];
  int N_minima;

  imagparams_t ret_params;

  int   *n_minima;
  int   **x_minima;
  int   **y_minima;
  float **err_minima;

  /* Use all the global values for the variables from sparms */
  nfeatures      = sparms.nfeatures;
  templatesize   = sparms.templatesize;
  template       = sparms.template;
  N_dim          = sparms.N_DIM;
  M_dim          = sparms.M_DIM;
  sets           = sparms.sets;
  bytes_pixel    = sparms.bytes_pixel;
  ncol           = SCALEWARP_NCOL;
  nrow           = SCALEWARP_NROW;

  max_N          = sparms.max_N;
  do_graymap     = sparms.do_graymap;
  do_minima      = sparms.do_minima;
  eig_first      = sparms.eig_first;
  eig_last       = sparms.eig_last;
  do_correlation = sparms.do_correlation;
  do_mahalanobis = sparms.do_mahalanobis;
  do_unitmap     = sparms.do_unitmap;
  do_spatial     = sparms.do_spatial;
  do_graymap     = sparms.do_graymap;

  if (do_correlation)  {   /* if norm_corr then you can't also do mahalanobis */
    do_mahalanobis = 0;
    do_unitmap = 0;
    do_spatial = 0;
  }

  /* ---- determine left/right boundaries for errormaps ----- */
  for (i=1; i<=nfeatures; i++) {
    rowl[i] = (int) templatesize[i][1]/2.0;
    rowr[i] = templatesize[i][1] - rowl[i] - 1;
    coll[i] = (int) templatesize[i][2]/2.0;
    colr[i] = templatesize[i][2] - coll[i] - 1;
  }

  patch = vector(1, max_N);
  my_vector = vector(1, MAX_VECTOR_LENGTH);

timestamp("eig_search: start");

  /* --- allocate errormap tensor and set window boundaries --- */

  for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
    errormap[k] = matrix(1, nrow, 1, ncol);
  
  if (DO_WINDOWS)
    {
      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
	{
	  imin[k] = w[k][1];
	  imax[k] = w[k][2];
	  jmin[k] = w[k][3];
	  jmax[k] = w[k][4];
	}
    }
  else 
    {
      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
	{
	  imin[k] = rowl[k] + 1;
	  imax[k] = nrow - rowr[k];
	  jmin[k] = coll[k] + 1;
	  jmax[k] = ncol - colr[k];
	}
    }
  
  if (do_minima) {
    int i1 = imin[feature], i2 = imax[feature];
    int j1 = jmin[feature], j2 = jmax[feature];
    int numpixels = (i2-i1)*(j2-j1), n;
    
    for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) {
      n = (imax[k]-imin[k])*(jmax[k]-jmin[k]);
      if (n>numpixels) {
	numpixels = n;
	i1 = imin[k];
	i2 = imax[k];
	j1 = jmin[k];
	j2 = jmax[k];
      }
    }
    N_minima = (i2-i1)*(j2-j1) / 9;
    *num_minima = N_minima;
    for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) {
      myPoints_array[k].n = N_minima;
      myPoints_array[k].x = ivector(1, N_minima);
      myPoints_array[k].y = ivector(1, N_minima);
      myPoints_array[k].f =  vector(1, N_minima);
    }
  }
  
  n_minima   = ivector(1, 4);
  x_minima   = imatrix(1, 4, 1, N_minima);
  y_minima   = imatrix(1, 4, 1, N_minima);
  err_minima = matrix(1, 4, 1, N_minima);

  /* --- check that specified template sizes and windows match image --- */

  for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
    if ((imin[k]-rowl[k]<1) || (imax[k]+rowr[k]>nrow) || \
	(jmin[k]-coll[k]<1) || (jmax[k]+colr[k]>ncol)) {
      fprintf(stderr,
	      "ERROR: image dimensions %d-by-%d don't match templates\n",
	      nrow, ncol);
      exit(3);
    }    
  
  
  /* compute mean template statistics for normalized correlation */
  
  if (do_correlation) {
    
    for (i=1; i<=nfeatures; i++) {
      statistics(template[i][1], N_dim[i],
		 &template_mean[i], &template_sigma[i]);
      fprintf(stdout,
	      "Normalized Correlation: Template-%d \t mean = %f \t sigma = %f\n",
	      i, template_mean[i], template_sigma[i]);
    }
  }
  
  
  /* ------------------------------------------------------ */
  /* -------------- MAIN Sequence Loop -------------------- */
  /* ------------------------------------------------------ */
    
  nframe = 1;
  
  /* ---- read current frame ----- */
  
  /* ------- initialize error map ----- */
  
  for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++)
    for (i=1; i<=nrow; i++)
      for (j=1; j<=ncol; j++)
	errormap[k][i][j] = 0.0;
  
  /* ------- process current frame -------- */
  
  for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) {

timestamp("eig_search: feature loop");
    
    fval = FLT_MYMAX;
    for (i=imin[k]; i<=imax[k]; i++)
      for (j=jmin[k]; j<=jmax[k]; j++) {
	
	c = 0;
	for (jj=j-coll[k]; jj<=j+colr[k]; jj++)
	  for (ii=i-rowl[k]; ii<=i+rowr[k]; ii++) 
	    patch[++c] = sparms.orig_image[ii][jj];
	
	if (do_unitmap) {
	  statistics(patch, N_dim[k], &mean, &sigma);
	  fval2 = umap_musigma_distance(mean, sigma,
					 umap_musigma[k]);
	}

	if (do_mahalanobis) {
	  
	  fval1 = mahalanobis(patch, N_dim[k], template[k], 
			      variances[k], eig_first, eig_last);
	  
	}
	else if (do_correlation) {
	  
	  fval1 = 1.0 - norm_corr(patch, N_dim[k], template[k][1],
				  template_mean[k], template_sigma[k]);
	}
	else
	  fval1 = dffs(patch,N_dim[k],template[k],eig_first,eig_last);
	
	if (do_unitmap) 
	  fval1 += fval2;

	  
	if (do_spatial) {
	  my_vector[1] = i;
	  my_vector[2] = j;
	  fval1 +=
	    mahalanobis_distance(my_vector, 2,
				 spatial_musigma[k][1],
				 &spatial_musigma[k][1]);
	}
	
	errormap[k][i][j] = fval1;
	if (fval1<fval) {
	  fval = fval1;
	  rowm[k] = i;
	  colm[k] = j;
	  error[k] = fval;
	}
	
      }


    if (do_minima) {
      minima(errormap[k], imin[k], imax[k], jmin[k], jmax[k],
	     &myPoints_array[k]);
      n_minima[k] = myPoints_array[k].n;
      for (i=1; i<=n_minima[k]; i++) {
	x_minima[k][i]   = myPoints_array[k].x[i];
	y_minima[k][i]   = myPoints_array[k].y[i];
	err_minima[k][i] = myPoints_array[k].f[i];
      }
    }

    /* --- DEBUG  dumpe k=1 errormap for bc3.ppm --- */
    {
      if (n_minima[k] == 0)  {
	fprintf(stderr,"\nWARNING: No minima found for feature %d !!!!\n\n", k);
        write_BIN("./errormap.bf", errormap[k], 128, 128);
	n_minima[k]      = 1;
	x_minima[k][1]   = rowm[k];
	y_minima[k][1]   = colm[k];
	err_minima[k][1] = error[k];
      }
      
    }



    
  } /* end of feature sequence loop */  

  *n_min   = n_minima;
  *x_min   = x_minima;
  *y_min   = y_minima;
  *err_min = err_minima;

timestamp("eig_search: done");

  /* --- free up allocated matrices ----- */
  
  free_vector(my_vector, 1, MAX_VECTOR_LENGTH);
  free_vector(patch, 1, max_N);
 
  for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
    free_matrix(errormap[k], 1, nrow, 1, ncol);

  return (ret_params);

}



