/*----------------------------------------------------------------------
PROGRAM: match.c
DATE:    07/11/94
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------

   Elastic Matching Interpretation Tree Energy Minimization code

	       
---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <float.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

/*----------------------------------------------------------------------*/

int enumerate(int *M, int N, int **P)

     /*  This function will enumerate all combinations of N 
	 objects each having M[i] prototypes and returns them 
	 in the matrix P[1..Ntotal][1..N] (pre-allocated) where 
	 each row is a combination
	 
	 It returns the total number of combinations

	 */

{

  register int i,j,k,l;
  int m1, m2, m3, mt, c;

  mt = 1;
  m1 = 1;
  for (i=1; i<=N; i++)
    mt *= M[i];

  for (i=1; i<=N; i++)  {

    m3 = m1;
    m1 = m1*M[i];
    m2 = mt/m1;
    c = 1;

    for (j=1; j<=m3; j++)
      for (k=1; k<=M[i]; k++)
	for (l=1; l<=m2; l++) 
	  P[c++][i] = k;

  }

  return mt;
}

/*----------------------------------------------------------------------*/

int best_match(struct Points *myPoints_array,
	       int **C_m, int M, int N,
	       float **Model,
	       float **K,
	       float mean_scale,
	       float **Match,
	       float *match_scale,
	       float *match_energy)


     /* This routine finds the minimum energy match between
	the feature positions in myPoints_array[1..N]
	and the model feature cood matrix Model[1..2][1..N]. 
	C_m[1..M][1..N] is the combination matrix.
	K[1..N][1..N] is the stiffness matrix.
	mean_scale is the expected average scale of the
	affine deformation.

	The routine return the index of the best combination
	(row in the matrix C_m), as well as the 3-by-N 
	matrix of best feature locations (where the 3rd
	row is the errors associated with each feature)
	as well as the scale and energy of the best match.

	*/

{

  register int i,j,c;
  float e, E, E_min, max_Dm;
  float **D, **Dm, **Data, **A, **temp;
  float my_scale, best_scale, v1, v2;
  int best_c = 1;


  D    = matrix(1, N, 1, N);
  Dm   = matrix(1, N, 1, N);
  Data = matrix(1, 3, 1, N);   /* note 3rd row is the distance */
  A    = matrix(1, 3, 1, 3);
  temp = matrix(1, 3, 1, 3);


  /* --- compute model distance matrix --- */
  for (i=1; i<=N; i++)
    for (j=i; j<=N; j++)  
      Dm[i][j] = Dm[j][i] = sqrt(SQR(Model[1][i]-Model[1][j]) + 
				 SQR(Model[2][i]-Model[2][j]));
  max_Dm = matrix_element_max(Dm, 1, N, 1, N);

  E_min = FLT_MAX;
  
  for (c=1; c<=M; c++) {  /* ---- loop thru all combinations ---*/
    
    
    /* ---- assign candidate points to Data matrix ---- */
    for (i=1; i<=N; i++) {
      Data[1][i] = myPoints_array[i].x[C_m[c][i]];
      Data[2][i] = myPoints_array[i].y[C_m[c][i]];
      Data[3][i] = myPoints_array[i].f[C_m[c][i]];
    }
    
    
    /* --- compute candidate distance matrix --- */
    for (i=1; i<=N; i++)
      for (j=i; j<=N; j++)  
	D[i][j] = D[j][i] = sqrt(SQR(Data[1][i]-Data[1][j]) +
				 SQR(Data[2][i]-Data[2][j]));
      
    
    /* --- compute scale change ---- */
    v1 = v2 = 0.0;
    for (i=1; i<=N; i++)
      for (j=i; j<=N; j++) {
	v1 += K[i][j] * Dm[i][j] * D[i][j];   
	v2 += K[i][j] * Dm[i][j] * Dm[i][j];  
      }
    my_scale = v1/v2;
    

    /* --- compute scale-compensated deformation energy  --- */
    e = 0;
    for (i=1; i<=N; i++)
      for (j=i+1; j<=N; j++)
	e += 0.5*K[i][j]*SQR(D[i][j] - my_scale*Dm[i][j]);  
    

    /* --- FERET energy function ----- */
    for (i=1; i<=N; i++) {
      Data[1][i] *= my_scale;   /* adjust the data for scale */
      Data[2][i] *= my_scale;   /* so later penalize only for shear */
    } 
    affine(Model, Data, N, A);
    if (A[1][1]*A[2][2]<0)
      E = FLT_MAX;            /* no xy-flips allowed */
    else {
      for (i=1; i<=3; i++)
	for (j=1; j<=3; j++)
	  temp[i][j] = A[i][j];
      for (i=1; i<=3; i++)
	temp[i][i] -= 1.0;
      v1 = (mean_scale)/my_scale;
      v2 = 1.0/v1;
      if (v1<v2) v1=v2;
      v2 = matrix_norm(temp, 2, 2);
      E = e + max_Dm * (SQR(SQR(v1))-1.0  + SQR(v2));
    }

    /* --- score/rank term ---- */
    v1 = 0;
    for (i=1; i<=N; i++) {
      v1 += Data[3][i]/myPoints_array[i].f[1];
    }
    E *= exp(v1-N);


    /* ---- update ---- */
    if (E<E_min) {
      best_c = c;
      best_scale = my_scale;
      E_min = E;
    }


    if (0) {
      
      fprintf(stdout,"combination: ");
      for (i=1; i<=N; i++)
	fprintf(stdout,"%3d ", C_m[c][i]);
      fprintf(stdout,"\n");

      fprintf(stdout,"features:\n");
      for (i=1; i<=2; i++) {
	for (j=1; j<=N; j++)
	  fprintf(stdout,"%3d ", (int) Data[i][j]);
	fprintf(stdout,"\n");
      }
      for (i=1; i<=2; i++) {
	for (j=1; j<=N; j++)
	  fprintf(stdout,"%3d ", (int) Model[i][j]);
	fprintf(stdout,"\n");
      }
      printf("Affine matrix:\n");
      for (i=1; i<=3; i++) {
	for (j=1; j<=3; j++)
	  fprintf(stdout,"%+3.2f\t",A[i][j]);
	fprintf(stdout,"\n");
      }

      fprintf(stdout,"scale = %f\ne = %f\nmax_Dm = %f\nv2 = %f\nenergy = %e\n",
	      my_scale,e,max_Dm,v2,E);
      fprintf(stdout,"\n\n");
    }

  }  /* --- end of combination loop ---- */
  


  /* ---- assign best candidate points to Match matrix ---- */


  for (i=1; i<=N; i++) {
    Match[1][i] = myPoints_array[i].x[C_m[best_c][i]];
    Match[2][i] = myPoints_array[i].y[C_m[best_c][i]];
    Match[3][i] = myPoints_array[i].f[C_m[best_c][i]];
  }
  *match_scale = best_scale;
  *match_energy = E_min;

 
  free_matrix(D, 1, N, 1, N);
  free_matrix(Dm, 1, N, 1, N);
  free_matrix(Data, 1, 3, 1, N);
  free_matrix(A, 1, 3, 1, 3);
  free_matrix(temp, 1, 3, 1, 3);
  
  return best_c;
}


/* ---------------------------------------------------------------------- */

int best_ml_match(struct Points *myPoints_array,
		  int **C_m, int M, int N,
		  float **Model,
		  float **Match,
		  float *Mean,       /* Mahalanobis mean  */
		  float **ICov,      /* Mahalanobis covar */
		  int Dim,           /* Mahalanobis dim   */
		  float *match_scale,
		  float *match_energy)


     /* This routine finds the maximum-likelihood match between
	the feature positions in myPoints_array[1..N]
	and the model feature cood matrix Model[1..2][1..N]. 
	C_m[1..M][1..N] is the combination matrix.
	K[1..N][1..N] is the stiffness matrix.
	mean_scale is the expected average scale of the
	affine deformation.

	The routine return the index of the best combination
	(row in the matrix C_m), as well as the 3-by-N 
	matrix of best feature locations (where the 3rd
	row is the errors associated with each feature)
	as well as the scale and energy of the best match.

	*/

{

  register int i,j,k,c;
  float E, E_min, max_Dm;
  float *V;
  float **D, **Dm, **Data, **A, **temp, **Theta;
  float my_scale, best_scale, v1, v2;
  int best_c = 1;
  float dx,dy;

  D    = matrix(1, N, 1, N);
  Theta= matrix(1, N, 1, N);

  Dm   = matrix(1, N, 1, N);
  Data = matrix(1, 3, 1, N);   /* note 3rd row is the distance */
  A    = matrix(1, 3, 1, 3);
  temp = matrix(1, 3, 1, 3);
  V    = vector(1, Dim);

  /* --- compute model distance matrix --- */
  for (i=1; i<=N; i++)
    for (j=i; j<=N; j++)  
      Dm[i][j] = Dm[j][i] = sqrt(SQR(Model[1][i]-Model[1][j]) + 
				 SQR(Model[2][i]-Model[2][j]));
  max_Dm = matrix_element_max(Dm, 1, N, 1, N);


  E_min = FLT_MAX;
  
  for (c=1; c<=M; c++) {  /* ---- loop thru all combinations ---*/
    
    
    /* ---- assign candidate points to Data matrix ---- */
    for (i=1; i<=N; i++) {
      Data[1][i] = myPoints_array[i].x[C_m[c][i]];
      Data[2][i] = myPoints_array[i].y[C_m[c][i]];
      Data[3][i] = myPoints_array[i].f[C_m[c][i]];
    }

    
    /* --- compute candidate distance/angle matrices --- */
    for (i=1; i<=N; i++)
      for (j=i+1; j<=N; j++)  {
	dx = Data[1][j] - Data[1][i];
	dy = Data[2][j] - Data[2][i];
	D[i][j] = D[j][i] = sqrt(dx*dx + dy*dy);
	if (dx==0 && dy==0) 
	  Theta[i][j] = 1000000000.0;  /* a big number */
	  /*** MAS: causes linux error: Theta[i][j] = FLT_MAX; ***/
	else
	  Theta[i][j] = atan2(dy,dx);
	Theta[j][i] = Theta[i][j];
      }


    /* --- compute scale change ---- */
    v1 = v2 = 0.0;
    for (i=1; i<=N; i++)
      for (j=i; j<=N; j++) {
	v1 += Dm[i][j] * D[i][j];   
	v2 += Dm[i][j] * Dm[i][j];  
      }
    my_scale = v1/v2;


    /* --- set-up geometry vector  --- */
    k = 1;

    for (i=1; i<=N; i++)
      for (j=i+1; j<=N; j++) 
	V[k++] = D[i][j];

    for (i=1; i<=N; i++)
      for (j=i+1; j<=N; j++) 
	V[k++] = Theta[i][j];

    /* ---------------------    
    affine(Model, Data, N, A);
    V[k++] = A[1][1];
    V[k++] = A[2][1];
    V[k++] = A[1][2];
    V[k++] = A[2][2];
    ---------------------- */

    if (k-1 != Dim) {
      fprintf(stderr,
	      "MATCH: Geometry vector dimension error %d != %d\n", k-1, Dim);
      exit(10);
    }

    
    /* --- compute Mahalanobis distance --- */

    E = local_mahalanobis_distance(V, Dim, Mean, ICov);


    /* --- score/rank term ---- */

    /* ------ SKIP --------
    v1 = 0;
    for (i=1; i<=N; i++) {
      v1 += Data[3][i]/myPoints_array[i].f[1];
    }
    E *= exp(v1-N);
    ---------------------- */
    v1 = 0;
    for (i=1; i<=N; i++) {
      v1 += Data[3][i];
    }
    E = E + v1;  


    /* ---- update ---- */
    if (E<E_min) {
      best_c = c;
      best_scale = my_scale;
      E_min = E;
    }


    if (0) {
      

      fprintf(stdout,"combination: ");
      for (i=1; i<=N; i++)
	fprintf(stdout,"%3d ", C_m[c][i]);
      fprintf(stdout,"\n");

      fprintf(stdout,"Model/Data:\n");
      for (i=1; i<=2; i++) {
	for (j=1; j<=N; j++)
	  fprintf(stdout,"%3d ", (int) Model[i][j]);
	fprintf(stdout,"\n");
      }
      for (i=1; i<=2; i++) {
	for (j=1; j<=N; j++)
	  fprintf(stdout,"%3d ", (int) Data[i][j]);
	fprintf(stdout,"\n");
      }
      printf("Affine matrix:\n");
      for (i=1; i<=3; i++) {
	for (j=1; j<=3; j++)
	  fprintf(stdout,"%+3.4f\t",A[i][j]);
	fprintf(stdout,"\n");
      }

      fprintf(stdout,"scale = %f\nenergy = %e\n",
	      my_scale,E);
      fprintf(stdout,"\n\n");
    }

  }  /* --- end of combination loop ---- */
  


  /* ---- assign best candidate points to Match matrix ---- */


  for (i=1; i<=N; i++) {
    Match[1][i] = myPoints_array[i].x[C_m[best_c][i]];
    Match[2][i] = myPoints_array[i].y[C_m[best_c][i]];
    Match[3][i] = myPoints_array[i].f[C_m[best_c][i]];
  }
  *match_scale = best_scale;
  *match_energy = E_min;

  free_vector(V, 1, Dim);
  free_matrix(D, 1, N, 1, N);
  free_matrix(Theta, 1, N, 1, N);
  free_matrix(Dm, 1, N, 1, N);
  free_matrix(Data, 1, 3, 1, N);
  free_matrix(A, 1, 3, 1, 3);
  free_matrix(temp, 1, 3, 1, 3);
  
  return best_c;
}


/* ---------------------------------------------------------------------- */

float local_mahalanobis_distance(float *x, int N, float *Mu, float **inv_Sigma)

/*
  This function returns the Mahalanobis distance for a feature vector
*/

{
  register int i,j;
  float sum=0, distance=0;
  float x_n[MAX_VECTOR_LENGTH];

  for (i=1; i<=N; i++)
    x_n[i] = x[i] - Mu[i];

  for (i=1; i<=N; i++) {
    sum = 0;
    for (j=1; j<=N; j++) {

      /*** MAS: DEBUG
      if (inv_Sigma[i][j] < -1000000.0 || inv_Sigma[i][j] > 1000000.0 ||
          x_n[j] < -1000000.0 || x_n[j] > 1000000.0) {
        printf("oldsum=%f; inv_Sigma[%d][%d]=%f * x_n[%d]=%f\n",
                sum, i,j, inv_Sigma[i][j], j, x_n[j]);
        printf("x[%d]=%f; Mu[%d]=%f\n", j, x[j], j, Mu[j]);
        fflush(stdout);
      } 
      ***/

      sum += inv_Sigma[i][j]*x_n[j];
    }

    distance += x_n[i]*sum;
  }

  return distance;
}

