/*----------------------------------------------------------------------
PROGRAM: ml_detect.c
DATE:    7/11/94
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------


  Feature Geometry Validation using Maximum Likelihood Minimization

  This routine uses data from BF format files: dmf%dt%.bf from the indir
  and a feature model matrix (in BF) format as well as a geometry_musigma.bf
  file

  And dumps the coods of best feature matches in output data structure

---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <limits.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

imagparams_t ml_detect(srchparams_t sparams, int **x_min, int **y_min,
		       float **err_min, int *Nmin)
{

  register int i,j,k;
  int nframe, nfeatures, sets, bytes_pixel;
  int nrow, ncol, image_nrow, image_ncol;

  int i_min, i_max, j_min, j_max, N_minima;

  struct Points myPoints_array[MAX_NUM_TEMPLATES];

  int top_N[MAX_NUM_TEMPLATES];
  int **Combination_matrix;
  int num_combinations_max;
  int num_combinations;
  int best_c;
  float **Model;
  float **Match;
  float **Stiffness;
  float match_scale, match_energy;

  int Geom_Dim;
  float *Geom_Mean;
  float **Geom_Cov;
  float **Geom_ICov;

  /* command line defaults */

/*  float mean_scale = MEAN_SCALE; */
  int top_N_max = TOP_N_MAX;
  int do_read_minima = DO_READ_MINIMA;    

  imagparams_t ret_iparams;
  
  sets = sparams.sets;
  bytes_pixel = sparams.bytes_pixel;
  image_ncol = SCALEWARP_NCOL;
  image_nrow = SCALEWARP_NROW;
  
  if (sets>1) 
    myerror("Input files must be single-set DAT files!");

  /* ---- read model file ----- */
  Model = sparams.model;
  nrow  = sparams.max_N;
  ncol  = sparams.nfeatures;
  if (nrow<2) {
    fprintf(stderr,"ERROR: Model file must have 2 rows!\n\n");
    exit(1);
  }
  nfeatures = ncol;


  /* ---- read geometry info ----- */
  { 
    float **M;

    nrow = sparams.geo_nrow;
    ncol = sparams.geo_ncol;
    M    = sparams.geo_M;

    Geom_Dim = 0;
    for (i=1; i<= nfeatures; i++)
      for (j=i+1; j<=nfeatures; j++)
	Geom_Dim++;
    Geom_Dim = 2 * Geom_Dim;             /* inter-dist + inter-angles */

    if (ncol != Geom_Dim) {
      fprintf(stderr,"ERROR: Geometry file must have %d cols not %d!\n\n",
	      Geom_Dim, ncol);
      exit(2);
    }
    if (nrow != Geom_Dim+1) {
     fprintf(stderr,"ERROR: Geometry file must have %d rows not %d!\n\n",
	      Geom_Dim+1, nrow);
      exit(2);
    }
    
    Geom_Mean = vector(1, Geom_Dim);
    Geom_Cov  = matrix(1, Geom_Dim, 1, Geom_Dim);
    Geom_ICov = matrix(1, Geom_Dim, 1, Geom_Dim);
    
    for (i=1; i<=Geom_Dim; i++)
      Geom_Mean[i] = M[1][i];

    for (i=1; i<=Geom_Dim; i++)
      for (j=1; j<=Geom_Dim; j++)
	Geom_Cov[i][j] = M[i+1][j];

    matrix_inverse(Geom_Cov, Geom_Dim, Geom_ICov);

  }



  /* ---- now determine max # of minima and allocate space ------ */

  N_minima = MAX_NUM_MINIMA;
  if (do_read_minima == 0) {
    i_max = image_nrow;
    j_max = image_ncol;
    i_min = j_min = 1;
    N_minima = (i_max - i_min) * (j_max - j_min) / 9;
  }


  for (k=1; k<=nfeatures; k++) {
    myPoints_array[k].n = N_minima;
    myPoints_array[k].x = ivector(1, N_minima);
    myPoints_array[k].y = ivector(1, N_minima);
    myPoints_array[k].f =  vector(1, N_minima);
  }



  /* ----  set up minimization stuff ---- */

  Stiffness = matrix(1, nfeatures, 1, nfeatures);
  for (i=1; i<=nfeatures; i++)
    for (j=1; j<=nfeatures; j++)
      Stiffness[i][j] = 1.0;

  num_combinations_max = 1;
  for (i=1; i<=nfeatures; i++) {
    top_N[i] = top_N_max;
    num_combinations_max *= top_N_max;
  }
  Combination_matrix = imatrix(1, num_combinations_max, 1, nfeatures);
  Match = matrix(1, 3, 1, nfeatures);

  /* ------------------------------------------------------ */
  /* -------------- MAIN Sequence Loop -------------------- */
  /* ------------------------------------------------------ */
 
  nframe = 1;
      
  /* ---- find/read the local minima in each map -------- */
      
  if (do_read_minima) 
    for (k=1; k<=nfeatures; k++) {
      myPoints_array[k].n = Nmin[k];
      for (i=1; i<= myPoints_array[k].n; i++){
	myPoints_array[k].x[i] = x_min[k][i];
	myPoints_array[k].y[i] = y_min[k][i];
	myPoints_array[k].f[i] = err_min[k][i];
      }
    }
  else 
    for (k=1; k <= nfeatures; k++) 
      minima(sparams.orig_image, 1, image_nrow, 1, image_ncol, &myPoints_array[k]);
  
      
  
  /* ---- minimzation search ------- */
  
  
  for (k=1; k<=nfeatures; k++)
    if (myPoints_array[k].n<top_N_max)
      top_N[k] = myPoints_array[k].n;
  i = 1;
  for (k=1; k<=nfeatures; k++)   /* this is superfluous, cuz inside */
    i *= top_N[k];               /* enumerate, the same is computed */
  
  num_combinations = enumerate(top_N,
			       nfeatures,
			       Combination_matrix);
  
  best_c = best_ml_match(myPoints_array,
			 Combination_matrix, num_combinations, nfeatures,
			 Model,
			 Match,
			 Geom_Mean,
			 Geom_ICov,
			 Geom_Dim,
			 &match_scale,
			 &match_energy);
 
  ret_iparams.lefteye_x  = (int) Match[1][1];
  ret_iparams.lefteye_y  = (int) Match[2][1];
  ret_iparams.righteye_x = (int) Match[1][2];
  ret_iparams.righteye_y = (int) Match[2][2];
  ret_iparams.nose_x     = (int) Match[1][3];
  ret_iparams.nose_y     = (int) Match[2][3];
  ret_iparams.mouth_x    = (int) Match[1][4];
  ret_iparams.mouth_y    = (int) Match[2][4];
  ret_iparams.energy     = match_energy;
    
  /* --- free up allocated matrices ----- */


  for (k=1; k<=nfeatures; k++) {
    free_ivector(myPoints_array[k].x, 1, N_minima);
    free_ivector(myPoints_array[k].y, 1, N_minima);
    free_vector(myPoints_array[k].f, 1, N_minima);
  }

  free_matrix(Stiffness, 1, nfeatures, 1, nfeatures); 
  free_imatrix(Combination_matrix, 1, num_combinations_max, 1, nfeatures);
  free_matrix(Match, 1, 3, 1, nfeatures);  

  free_vector(Geom_Mean, 1, Geom_Dim);
  free_matrix(Geom_Cov,  1, Geom_Dim, 1, Geom_Dim);
  free_matrix(Geom_ICov, 1, Geom_Dim, 1, Geom_Dim);
  
  return (ret_iparams);

}



/*------------------------------------------------------------------- */
/* -------------------- end of main() ------------------------------- */
/*------------------------------------------------------------------- */
