/*----------------------------------------------------------------------
PROGRAM: modelwarp.c
DATE:    6/2/94
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------


   This contains code to warp images according to the affine transform
   between the feature locations indicated and those in 
   the model data structure (from the model.bf file defined in defines.h)

---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <float.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

void model_warp(imagparams_t iparams, srchparams_t sparams, image_t ret_image)
{

  register int i;
  int nframe, nfeatures, sets, bytes_pixel;
  int nrow, ncol, M;

  float **P, **Q, **W; 

  int rigid_select = DO_RIGID;
  int rigid_pt_1 = DO_RIGID_PT1;
  int rigid_pt_2 = DO_RIGID_PT2;

  sets = sparams.sets;
  bytes_pixel = sparams.bytes_pixel;
  ncol = SCALEWARP_NCOL;
  nrow = SCALEWARP_NROW;
  nfeatures = sparams.nfeatures;
  M = sparams.max_N;

  if (sets>1) 
    myerror("Input files must be single-set DAT files!");

  P = sparams.model;
  Q = matrix(1, M, 1, nfeatures);
  W    = matrix(1, 3, 1, 3);

  nframe = 1;
  Q[1][1] = iparams.lefteye_x;
  Q[1][2] = iparams.righteye_x;
  Q[1][3] = iparams.nose_x;
  Q[1][4] = iparams.mouth_x;
  
  Q[2][1] = iparams.lefteye_y;
  Q[2][2] = iparams.righteye_y;
  Q[2][3] = iparams.nose_y;
  Q[2][4] = iparams.mouth_y;

  if (rigid_select == 1)
    rigid(Q, P, nfeatures, W, rigid_pt_1, rigid_pt_2);
  else
    affine(Q, P, nfeatures, W);
      
  /* --- apply warp --- */

  i = affine_warp(sparams.orig_image, ret_image, nrow, ncol, W, 0.0);
      
  free_matrix(Q, 1, M, 1, nfeatures);
  free_matrix(W, 1, 3, 1, 3);

  return;

} /* model_warp */

/*----------------------- MAS -----------------------------------------*/

/***
  This procedure operates as for model_warp() except that the return 
  image can be of different size from the original image (passed via
  sparams.orig_image and sparams.ncol/nrow parameters).

  MAS
***/

void model_warp_2(imagparams_t iparams, srchparams_t sparams, 
		  image_t ret_image, int ret_nrow, int ret_ncol)
{

  register int i,j;
  int nframe, nfeatures, sets, bytes_pixel;
  int nrow, ncol, M;
  int clipping = 0;
  image_t warped_im;

  float **P, **Q, **W;

  int rigid_select = DO_RIGID;
  int rigid_pt_1 = DO_RIGID_PT1;
  int rigid_pt_2 = DO_RIGID_PT2;

  sets = sparams.sets;
  bytes_pixel = sparams.bytes_pixel;
  ncol = sparams.ncol;    /* i/p image size */
  nrow = sparams.nrow;
  nfeatures = sparams.nfeatures;
  M = sparams.max_N;

  if (sets>1) 
    myerror("Input files must be single-set DAT files!");

  P = sparams.model;
  Q = matrix(1, M, 1, nfeatures);
  W    = matrix(1, 3, 1, 3);

  nframe = 1;
  Q[1][1] = iparams.lefteye_x;
  Q[1][2] = iparams.righteye_x;
  Q[1][3] = iparams.nose_x;
  Q[1][4] = iparams.mouth_x;
  
  Q[2][1] = iparams.lefteye_y;
  Q[2][2] = iparams.righteye_y;
  Q[2][3] = iparams.nose_y;
  Q[2][4] = iparams.mouth_y;

  if (rigid_select == 1)
    rigid(Q, P, nfeatures, W, rigid_pt_1, rigid_pt_2);
  else
    affine(Q, P, nfeatures, W);
      
  /* --- apply warp --- */

  /* i/p and o/p images are the same size */
  if (ret_nrow == nrow && ret_ncol == ncol)
    affine_warp(sparams.orig_image, ret_image, nrow, ncol, W, 0.0);

  else {

    if (ret_nrow > nrow || ret_ncol > ncol) {
      fprintf(stderr,"model_warp_2: return image too large\n");
      return;
    }
    
    warped_im = matrix(1, nrow, 1, ncol);
    affine_warp(sparams.orig_image, warped_im, nrow, ncol, W, 0.0);
      
    for (i=1; i < ret_nrow; i++)
      for (j=1; j < ret_ncol; j++)
	ret_image[i][j] = warped_im[i][j];
    
    free_matrix(warped_im, 1, nrow, 1, ncol);
  }
  
  free_matrix(Q, 1, M, 1, nfeatures);
  free_matrix(W, 1, 3, 1, 3);

  return;

} /* model_warp_2 */

