/*----------------------------------------------------------------------
PROGRAM: ms_eigsearch3.c
DATE:    10/14/93
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------


  Multiple-Scale Local feature search by eigentemplates

  This routine uses data from 2 (3) BF files (previously loaded in) 

  1. features.bf  which is an N-by-2 matrix defining the size (row,col)
     of each of the N templates.

  2. template[n].bf  where n=1...N is the eigenvector ROW matrix where
     the 1st row is the mean feature, and the remaining rows
     are the principal eigenvectors 1:M. 

     NOTE: The eigenvectors are stored in column-order (as in MATLAB)

  3. variances.bf which is an N-by-M matrix, where each row represents
     the rank-ordered eigenvalues associated with that feature


------------------------------------------------------------------------

   NOTE: In this version if unitmap'ing is selected it will use
         the Mahalanobis mean/stddev in addition to the standard
	 distance computation.

---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <float.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

extern void timestamp(char *label);

imagparams_t ms_search(srchparams_t sparms, 
		       srchparams_t EIG_sparams,
		       srchparams_t MLD_sparams,
		       imagparams_t *best_MLD_imgparams,
		       int hier_flag,
		       float **variances,
		       float **umap_musigma[MAX_NUM_TEMPLATES],
		       float **spatial_musigma[MAX_NUM_TEMPLATES],
		       float **eig_variances,
		       float **eig_umap_musigma[MAX_NUM_TEMPLATES],
		       float **eig_spatial_musigma[MAX_NUM_TEMPLATES],
		       float **w)
{

  register int i,j,k,ii,jj;
  int   s,c,feature=1, nfeatures, sets, bytes_pixel;
  int   nrow, ncol, max_N;

  float ***template;
  float **errormap[MAX_NUM_SCALES][MAX_NUM_TEMPLATES];
  
  int do_graymap = 0; /* default is NOT to graymap each patch */
  int multiple_search = 0; /* search for extra heads */
  int eig_first = 1;  /* search using first 5 eigenvectors by default */
  int eig_last  = 5;
  int do_minima; 
  int do_unitmap;
  int do_mahalanobis;
  int do_correlation;
  int do_spatial;

  float fval1, fval, fval2, mean, sigma;

  int   rowl[MAX_NUM_TEMPLATES],rowr[MAX_NUM_TEMPLATES];
  int   coll[MAX_NUM_TEMPLATES],colr[MAX_NUM_TEMPLATES];
  int   imin[MAX_NUM_TEMPLATES],imax[MAX_NUM_TEMPLATES];
  int   jmin[MAX_NUM_TEMPLATES],jmax[MAX_NUM_TEMPLATES];
  int   *N_dim; /* the dimensionality of each template */
  int   *M_dim; /* the # of eigenvectors for each template */

  /* --- FIX: these should really be defined in pipe_main --- */
  float template_mean[MAX_NUM_TEMPLATES];   /* statistics for norm_corr */
  float template_sigma[MAX_NUM_TEMPLATES];  /* statistics for norm_corr */

  int   numscales;
  float *scales;
  int   rowm[MAX_NUM_SCALES][MAX_NUM_TEMPLATES];
  int   colm[MAX_NUM_SCALES][MAX_NUM_TEMPLATES];
  float error[MAX_NUM_SCALES][MAX_NUM_TEMPLATES];
  float **Warp;
  float *my_vector;
  int   xdim, ydim;
  int   best_scale[MAX_NUM_TEMPLATES];
  float best_error[MAX_NUM_TEMPLATES];
  int   best_rowm[MAX_NUM_TEMPLATES];
  int   best_colm[MAX_NUM_TEMPLATES];
  int   second_scale[MAX_NUM_TEMPLATES], third_scale[MAX_NUM_TEMPLATES];
  float second_error[MAX_NUM_TEMPLATES], third_error[MAX_NUM_TEMPLATES];
  int   second_rowm[MAX_NUM_TEMPLATES],  third_rowm[MAX_NUM_TEMPLATES];
  int   second_colm[MAX_NUM_TEMPLATES],  third_colm[MAX_NUM_TEMPLATES];
  

  int **templatesize;
  float *patch;
  
  int singlefeature = 1;

  imagparams_t  ret_params;
  image_t image;

  /* ---------------------- Hierarchical Stuff ----------------------*/

  imagparams_t     EIG_imgparams, MLD_imgparams;
  image_t          scaled_im;
  float            **ERR_minima;
  int              **X_minima, **Y_minima;
  int              *N_minima;
  int              rowm_list[MAX_NUM_TEMPLATES][HIER_TOP_N];
  int              colm_list[MAX_NUM_TEMPLATES][HIER_TOP_N];
  int              s_list[MAX_NUM_TEMPLATES][HIER_TOP_N];
  float            error_list[MAX_NUM_TEMPLATES][HIER_TOP_N];
  float            scale_list[MAX_NUM_TEMPLATES][HIER_TOP_N];
  

  /* ---------------------- End Hierarachical Stuff -----------------*/
  
  Warp = matrix(1, 3, 1, 3);
  scaled_im = matrix(1,SCALEWARP_NROW,1,SCALEWARP_NCOL);

  /* Use all the global values for the variables from sparms */

  nfeatures      = sparms.nfeatures;
  templatesize   = sparms.templatesize;
  template       = sparms.template;
  N_dim          = sparms.N_DIM;
  M_dim          = sparms.M_DIM;
  sets           = sparms.sets;
  bytes_pixel    = sparms.bytes_pixel;
  ncol           = sparms.ncol;
  nrow           = sparms.nrow;
  numscales      = sparms.numscales;
  max_N          = sparms.max_N;
  do_graymap     = sparms.do_graymap;
  multiple_search= sparms.multiple_search;
  eig_first      = sparms.eig_first;
  eig_last       = sparms.eig_last;
  scales         = sparms.scales;
  do_correlation = sparms.do_correlation;
  do_mahalanobis = sparms.do_mahalanobis;
  do_unitmap     = sparms.do_unitmap;
  do_spatial     = sparms.do_spatial;
  do_graymap     = sparms.do_graymap;

  if (do_correlation)  {   /* if norm_corr then you can't also do mahalanobis */
    do_mahalanobis = 0;
    do_unitmap = 0;  
    do_spatial = 0;
  }
    
  image = matrix (1, nrow, 1, ncol);
  do_minima = sparms.do_minima;
  
  /* ---- determine left/right boundaries for errormaps ----- */

  for (i=1; i<=nfeatures; i++) 
    {
      rowl[i] = (int) templatesize[i][1]/2.0;
      rowr[i] = templatesize[i][1] - rowl[i] - 1;
      coll[i] = (int) templatesize[i][2]/2.0;
      colr[i] = templatesize[i][2] - coll[i] - 1;
    }

  patch = vector(1, max_N);
  my_vector = vector(1, MAX_VECTOR_LENGTH);
  
  /* --- allocate errormap tensor and set window boundaries --- */

  for (s = 1; s <= numscales; s++)
    for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
      {
	errormap[s][k] = matrix(1, nrow, 1, ncol);
	imin[k] = rowl[k] + 1;
	imax[k] = nrow - rowr[k];
	jmin[k] = coll[k] + 1;
	jmax[k] = ncol - colr[k];
      }
  
  /* compute mean template statistics for normalized correlation */
  
  if (do_correlation) {
    
    for (i=1; i<=nfeatures; i++) {
      statistics(template[i][1], N_dim[i],
		 &template_mean[i], &template_sigma[i]);
      fprintf(stdout,
	      "Normalized Correlation: Template-%d \t mean = %f \t sigma = %f\n",
	      i, template_mean[i], template_sigma[i]);
    }
  }
  

  /* ------------------------------------------------------ */
  /* -------------- MAIN Sequence Loop -------------------- */
  /* ------------------------------------------------------ */

timestamp("ms_search:main sequence loop");
	
  for (s=1; s<=numscales; s++) 
    {

/*timestamp("ms_search:scale loop");*/

      /* ------- initialize error map ----- */
      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++)
	for (i=1; i<=nrow; i++)
	  for (j=1; j<=ncol; j++)
	    errormap[s][k][i][j] = 0.0; 
   
      /* ------ apply scaling to image_orig ------ */
      for (i=1; i<=3; i++)
	for (j=1; j<=3; j++)
	  Warp[i][j] = 0.0;
      Warp[1][1] = Warp[2][2] = scales[s];
      Warp[3][3] = 1.0;

/*timestamp("ms_search:affine warp");*/

      affine_warp(sparms.orig_image, image, nrow, ncol, Warp, 0.0);
      xdim = (int) (scales[s]*nrow + 0.5);
      ydim = (int) (scales[s]*ncol + 0.5);
 
      /* ----------------- Process current frame/scale --------------- */
	
      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
	{

/*timestamp("ms_search:feature loop");*/

	  imax[k] = xdim - rowr[k];
	  jmax[k] = ydim - colr[k];
      
      
	  fval = FLT_MYMAX;
	  for (i=imin[k]; i<=imax[k]; i++)
	    for (j=jmin[k]; j<=jmax[k]; j++) 
	      {	  
		c = 0;
		for (jj=j-coll[k]; jj<=j+colr[k]; jj++)
		  for (ii=i-rowl[k]; ii<=i+rowr[k]; ii++) 
		    patch[++c] = image[ii][jj];
	
		fval2 = 0;
		if (do_unitmap) 
		  {
		    statistics(patch, N_dim[k], &mean, &sigma);
		    fval2 = umap_musigma_distance(mean, sigma,
						   umap_musigma[k]);
		  }

	 	  
		if (do_mahalanobis) 
		  {
	    	    fval1 = mahalanobis(patch, N_dim[k], template[k], 
				variances[k], eig_first, eig_last);
		  }
		else if (do_correlation) {
		  
		  fval1 = 1.0 - norm_corr(patch, N_dim[k], template[k][1],
					  template_mean[k], template_sigma[k]);
		}
		else
		  fval1 = dffs(patch,N_dim[k],template[k],eig_first,eig_last);


		if (do_unitmap) 
		  fval1 += fval2;

		if (do_spatial) 
		  {
		    my_vector[1] = (int) (i/scales[s]+0.5);
		    my_vector[2] = (int) (j/scales[s]+0.5);
		    fval1 +=
		      mahalanobis_distance(my_vector, 2,
					   spatial_musigma[k][1],
					   &spatial_musigma[k][1]);
		  }
	      
		errormap[s][k][i][j] = fval1;

	      }


	  /* Get best minimum (check that it is a minimum!) */

	  if (!hier_flag)
	    best_minimum(imin[k],imax[k],jmin[k],jmax[k],
			 &rowm[s][k],&colm[s][k],&error[s][k],
			 errormap[s][k],scales[s]);

	} /* end of feature sequence loop */
    }  /* end of scale sequence loop */
  

  /* ---------------------- Hierarchical Stuff ----------------------

     This section of ms_search seeks to check to see if features are 
     in the right place for each candidate head location.  However,
     checking every single minimum is too time consuming, so this
     checks only one minimum in each HIER_BOX_SIZE area of image-scale
     space.  This way, only non-redundant head locations are checked.

     -------------------------------------------------------------------*/

  if (hier_flag)
    {
      int sl;              /* scale loop counter */
      int rl;              /* row loop counter */
      int cl;              /* column loop counter */
      int curr_s, curr_minrow, curr_maxrow, curr_mincol, curr_maxcol;
      int minrow, maxrow, mincol, maxcol;
      int n, rc, cc;       /* three more temporary counters */
      float curr_scale;
      int num_minima;
      int found_minimum;

timestamp("ms_search: hierarchical feature stuff");

      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
	{
	  best_error[k] = FLT_MYMAX;
	  for (n = 0; n < HIER_TOP_N-1; n++)
	    error_list[k][n] = FLT_MYMAX;
	}

      fval = FLT_MYMAX;
      found_minimum = FALSE;

      for (sl = 1; sl <= ceil(((float)numscales)/(float)HIER_BOX_SIZE); sl++) {

timestamp("ms_search: scale loop (hierarchical)");

	for (cl = 1; cl <= ceil(((float)ncol)/(float)HIER_BOX_SIZE); cl++)
	  for (rl = 1; rl <= ceil(((float)nrow)/(float)HIER_BOX_SIZE); rl++)
	    {
	      /* Calculate which pixels in each errormap correspond to */
	      /* the current region in the minima map.  Since the */
	      /* errormap pixels are for different scales, this */
	      /* calculation is necessary when making sure that only */
	      /* one minimum is checked in each HIER_BOX_SIZE region, */
	      /* so all the minima get projected up to the same image */
	      /* size and then only the best in the region is taken */
	      
	      mincol = (cl-1) * HIER_BOX_SIZE + 1;
	      maxcol = mincol + HIER_BOX_SIZE - 1;
	      minrow = (rl-1) * HIER_BOX_SIZE + 1;
	      maxrow = minrow + HIER_BOX_SIZE - 1;
	      for (n = 1; n <= HIER_BOX_SIZE; n++)
		{
		  curr_s = (sl-1) * HIER_BOX_SIZE + n;
		  if (curr_s <= numscales)
		    {
		      curr_scale = scales[curr_s];
		      curr_mincol = floor(mincol * curr_scale);
		      curr_maxcol = floor(maxcol * curr_scale);
		      curr_minrow = floor(minrow * curr_scale);
		      curr_maxrow = floor(maxrow * curr_scale);
		      
		      for (cc = curr_mincol; cc < curr_maxcol; cc++)
			for (rc = curr_minrow; rc < curr_maxrow; rc++)
			  for (k=feature; 
			       k<=(singlefeature>0 ? feature:nfeatures); k++)
			    if ((cc > 1) && (rc > 1) &&
				(cc < ncol) && (rc < nrow))
			      if ((check_minimum(errormap[curr_s][k],rc,cc))
				  && (errormap[curr_s][k][rc][cc] < fval))
				{
				  found_minimum = TRUE;
				  pop(error_list[k],rowm_list[k],colm_list[k],
				      scale_list[k],s_list[k],
				      errormap[curr_s][k][rc][cc],
				      rc,cc,curr_scale,curr_s);
				  fval = error_list[k][HIER_TOP_N-1];
				}
		    }
		}

	    }
      }

      n = 0;

      /* Since error_list is sorted, if the value at any n of */
      /* error_list is greater than the best_error, there is no point */
      /* in checking any members of error_list after that, so use a */
      /* while loop rather than a for loop */

      if (found_minimum)
	for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++)
	  while ((n < HIER_TOP_N) && (error_list[k][n] < best_error[k]))
	    {
	      if (error_list[k][n] < best_error[k])
		{

timestamp("ms_search: scale_warp (hierarchical)");

		  scale_warp(sparms,rowm_list[k][n],colm_list[k][n],
			     scale_list[k][n],scaled_im);
		  EIG_sparams.orig_image = scaled_im;
		  EIG_imgparams = eig_search(EIG_sparams, &X_minima,
					     &Y_minima, &ERR_minima,
					     &N_minima, eig_variances,
					     eig_umap_musigma,
					     eig_spatial_musigma, w,
					     &num_minima);
		  MLD_sparams.orig_image = scaled_im;
		  MLD_imgparams = ml_detect(MLD_sparams,X_minima,
					    Y_minima,ERR_minima,N_minima);
		  fprintf(stdout,
			  "\nAt scale %f, (%d,%d) has en: %f and MLD_en: %f\n",
			  scale_list[k][n],rowm_list[k][n],
			  colm_list[k][n],error_list[k][n],
			  MLD_imgparams.energy);
		  fprintf(stdout,
		  	  "Left eye X,Y = %d,%d; Right eye X,Y = %d,%d; "
		  	  "Nose X,Y = %d,%d; Mouth X,Y = %d,%d\n\n",
		  	  MLD_imgparams.lefteye_x, MLD_imgparams.lefteye_y,
		  	  MLD_imgparams.righteye_x, MLD_imgparams.righteye_y,
		  	  MLD_imgparams.nose_x, MLD_imgparams.nose_y,
		  	  MLD_imgparams.mouth_x, MLD_imgparams.mouth_y);
		  error_list[k][n] += MLD_imgparams.energy;
		  free_ivector(N_minima,1,4);
		  free_imatrix(X_minima,1,4,1,num_minima);
		  free_imatrix(Y_minima,1,4,1,num_minima);
		  free_matrix(ERR_minima,1,4,1,num_minima);
		  
		  if (error_list[k][n]<best_error[k]) 
		    {
		      best_error[k] = error_list[k][n];
		      best_scale[k] = s_list[k][n];
		      best_rowm[k]  = rowm_list[k][n];
		      best_colm[k]  = colm_list[k][n];
		      *best_MLD_imgparams = MLD_imgparams;
		    }
		  n++;
		}
	      }
      
      
    }
  
  /* --------------------- End Hierarchical Stuff -------------------*/

  else
    
    {
      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
	{ 
	  best_error[k] = FLT_MYMAX;
	  for (s=1; s<=numscales; s++) 
	    if (error[s][k]<best_error[k]) 
	      {
		best_error[k] = error[s][k];
		best_scale[k] = s;
		best_rowm[k]  = rowm[s][k];
		best_colm[k]  = colm[s][k];
	      }
	}
    }



/* MARK - HERE IS THE BEST ESTIMATE FOR X AND Y COORDINATES OF HEAD!!! */

  ret_params.best_head_x = best_rowm[1];
  ret_params.best_head_y = best_colm[1];
  fprintf(stdout,"Best Head x:  %d Best Head y:  %d  Best Scale:  %f\n",
	  best_rowm[1],best_colm[1],scales[best_scale[1]]);
  ret_params.best_head_e = best_error[1];
  ret_params.best_scale  = scales[best_scale[1]];
	     
  if (multiple_search)
    {
      for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++){
	second_error[k] = FLT_MYMAX;
	for (s=1; s<=numscales; s++)
	  if ((error[s][k] < second_error[k]) && (error[s][k] > best_error[k]))
	    {
	      second_error[k] = error[s][k];
	      second_scale[k] = s;
	      second_rowm[k]  = rowm[s][k];
	      second_colm[k]  = colm[s][k];
	    }
	third_error[k] = FLT_MYMAX;
	for(s=1; s<=numscales; s++)
	  if((error[s][k] < third_error[k]) && (error[s][k] > second_error[k]))
	    {
	      third_error[k] = error[s][k];
	      third_scale[k] = s;
	      third_rowm[k]  = rowm[s][k];
	      third_colm[k]  = colm[s][k];
	    }
      }
   
      printf("2nd best head x = %d,2nd best head y = %d 2nd best scale = %f\n",
	     second_rowm[1], second_colm[1], scales[second_scale[1]]);
  
      ret_params.second_head_y = second_colm[1];
      ret_params.second_head_x = second_rowm[1];
      ret_params.second_head_e = second_error[1];
      ret_params.second_scale  = scales[second_scale[1]];
      
      printf("3rd best head x = %d,3rd best head y = %d 3rd best scale = %f\n",
	     third_rowm[1], third_colm[1], scales[third_scale[1]]);
      ret_params.third_head_y  = third_colm[1];
      ret_params.third_head_x  = third_rowm[1];
      ret_params.third_head_e  = third_error[1];
      ret_params.third_scale   = scales[third_scale[1]];
    }
  /* --- free up allocated matrices ----- */
  
timestamp("ms_search: done");

  for (s = 1; s <= numscales; s++)
    for (k=feature; k<=(singlefeature>0 ? feature:nfeatures); k++) 
      free_matrix(errormap[s][k], 1, nrow, 1, ncol);

  free_vector(my_vector, 1, MAX_VECTOR_LENGTH);
  free_vector(patch, 1, max_N);
  free_matrix(image, 1, nrow, 1, ncol);
  free_matrix(Warp, 1, 3, 1, 3);
  free_matrix(scaled_im,1,SCALEWARP_NROW,1,SCALEWARP_NCOL);
  
  return(ret_params);

  
}



/*------------------------------------------------------------------- */
/* -------------------- end of main() ------------------------------- */
/*------------------------------------------------------------------- */

/*------------------------------------------------------------------- */

float norm_corr(float *patch, int N,
		float *template, float t_mean, float t_sigma)
{
  register int i;
  float sum,mean,sigma;

  statistics(patch, N, &mean, &sigma);

  sum = 0;
  for (i=1; i<=N; i++)
    sum += patch[i]*template[i];

  return (sum/N - t_mean*mean)/(t_sigma*sigma);
}

/* ----------------------------------------------------------------------*/


float dffs(float *patch, int N, float **eigvectors, int first, int last)
{
  register int i,k;
  float error,sum,mean,var=1.0,E,Ec,Em;
  int M = 1;
  
  static float C[MAX_NUM_EIGENVECTORS];


  if (first>0)
    M = last - first + 1;
  /* compute the first 2 moments of image patch */
 
  if (DO_VAR) {
    statistics(patch, N, &mean, &var);
    var = SQR(var);
  }


  /* graymap/unitmap the patch */

  if (DO_GRAYMAP)  
    graymap(patch, N);
  else if (DO_UNITMAP)
    unitmap(patch, N);
  
  /* subtract the mean from the patch (note mean is first row)
     and compute the Energy of the patch                       */

  E = 0.0;
  for (i=1; i<=N; i++) {
    patch[i] -= eigvectors[1][i];
    E += SQR(patch[i]);
  }


  /* project onto the appropriate eigenvectors */
  
  Ec = Em = 0.0;   /* if first=0 then this is just mean-template SSD */
  if (first>0) {   
    for (k = first+1; k<=last+1; k++) {
      sum = 0.0;
      for (i=1; i<=N; i++)
	sum += patch[i] * eigvectors[k][i];
      C[k] = sum;
      Ec += SQR(C[k]);
    }
  }

  error = (E - Ec);
  if (DO_VAR)
    if (var<VAR_THRESHOLD)
      error = VAR_MAXVALUE;


  return error;
}

/* ---------------------------------------------------------------------- */

float mahalanobis(float *patch, int N, float **eigvectors, float *eigvalues,
		  int first, int last)
{
  register int i;
  register float *pv1,*pv2,*pv3;
  int k;
  float error,sum,mean,var=1.0,sigma,E,Ec,Em,val;
  int M = 1;
  
  static float C[MAX_NUM_EIGENVECTORS];


  if (first>0)
    M = last - first + 1;
  
  /* compute the first 2 moments of image patch */
  
  if (DO_VAR) {
    statistics(patch, N, &mean, &sigma);
    var = SQR(sigma);
  }

  /* graymap/unitmap the patch */

  if (DO_GRAYMAP)  
    graymap(patch, N);
  else if (DO_UNITMAP)
    unitmap(patch, N);

  /* subtract the mean from the patch (note mean is first row)
     and compute the Energy of the patch                       */

  E = 0.0;
/*
  for (i=1; i<=N; i++) {
    patch[i] -= eigvectors[1][i];
    E += SQR(patch[i]);
  }
*/
  pv1 = patch+1;
  pv2 = eigvectors[1]+1;
  for (i=1; i<=N; i++) {
    val = (*pv1++ -= *pv2++);
    E += SQR(val);
  }

  /* project onto the appropriate eigenvectors */
  
  Ec = Em = 0.0;   /* if first=0 then this is just mean-template SSD */

/*
  if (first>0) {   
    for (k = first+1; k<=last+1; k++) {
      sum = 0.0;
      for (i=1; i<=N; i++)
	sum += patch[i] * eigvectors[k][i];
      C[k] = sum;
      val = SQR(C[k]);
      Ec += val;
      Em += val/eigvalues[k-1];
    }
  }
*/
  if (first>0) {   
    pv3 = C;
    for (k = first+1; k<=last+1; k++) {
      sum = 0.0;
      pv1 = patch+1;
      pv2 = eigvectors[k]+1;
      for (i=1; i<=N; i++)
	sum += *pv1++ * *pv2++;
      *pv3++ = sum;
      val = SQR(sum);
      Ec += val;
      Em += val/eigvalues[k-1];
    }
  }

  if (DO_VAR)
    if (var<VAR_THRESHOLD)
      error = VAR_MAXVALUE;

  error = Em + (E - Ec)/eigvalues[last+1];
  return error;
}


/*---------------------------------------------------------------------- */


float graymap(float *p, int N)

/* Maps the N-dimensional float vector p from its min/max to 0/255 */

{

  register int i;
  float vmin,vmax,range,scale;

  vmin = FLT_MAX;
  vmax = -vmin;

  for (i=1; i<=N; i++) {
    if (p[i]<vmin)
      vmin = p[i];
    if (p[i]>vmax)
      vmax = p[i];
  }

  range = vmax-vmin;
  if (range==0.0)         /* if no variation don't modify array */
    return 0;

  scale = 255.0/range;

  for (i=1; i<=N; i++)
    p[i] = (int) (0.5 + scale*(p[i]-vmin));

  return range;
}



/*---------------------------------------------------------------------- */


float unitmap(float *p, int N)

/* Maps the N-dimensional float vector p to zero-mean/unit-std dev. */
/* It returns the standard deviation (or zero if sigma<FLT_MIN).    */

{

  register int i;
  register float *pv1, *pv2;
  float mean = 0, sigma = 0;

  statistics(p, N, &mean, &sigma);

/*
  if (sigma>FLT_MIN) {
    for (i=1; i<=N; i++)
      p[i] = (p[i]-mean)/sigma;
    return sigma;
  }
  else {
    for (i=1; i<=N; i++)
      p[i] -= mean;
    return 0;
  }
*/
  if (sigma>FLT_MIN) {
    for (i=1, pv1=p+1; i<=N; i++, pv1++)
      *pv1 = (*pv1-mean)/sigma;
    return sigma;
  }
  else {
    for (i=1, pv1=p+1; i<=N; i++)
      *pv1++ -= mean;
    return 0;
  }
}


/* ---------------------------------------------------------------------- */

void statistics(float *p, int N, float *mean, float *sigma)

/* computes the mean and standard deviation of the vector p */

{
  register int i;
  register float *pv1, *pv2;
  float val, mymean = 0, mysigma = 0; 
  
  pv1 = p+1;
  for (i=1; i<=N; i++)
    mymean += *pv1++;
  mymean /= N;

  pv1 = p+1;
  for (i=1; i<=N; i++) {
    val = *pv1++ - mymean;
    mysigma += val*val;
  }
  mysigma = sqrt(mysigma/(N-1));

  *mean = mymean;
  *sigma = mysigma;
}


/* ---------------------------------------------------------------------- */

float umap_musigma_distance(float mean, float sigma, float **umap_musigma)

/* This function takes the mean and sigma of the patch and returns
   the Mahalanobis distance according to the 2nd-order statistics
   in the matrix umap_musigma (whose first row is the mean, and 2nd-3rd
   the INVERSE covariance
*/

{
  register int i,j;
  float sum=0, distance=0;
  float x[3],y[3];

  x[1] = mean  - umap_musigma[1][1];
  x[2] = sigma - umap_musigma[1][2];

  for (i=1; i<=2; i++) {
    sum = 0;
    for (j=1; j<=2; j++)
      sum += umap_musigma[i+1][j]*x[j];
    y[i] = sum;
  }

  distance = 0;
  for (i=1; i<=2; i++)
    distance += x[i]*y[i];

  return distance;
}

/* ---------------------------------------------------------------------- */

float mahalanobis_distance(float *x, int N, float *Mu, float **inv_Sigma)

/*
  This function returns the Mahalanobis distance for a feature vector
*/

{
  register int i,j;
  float sum=0, distance=0;
  float x_n[MAX_VECTOR_LENGTH];

  for (i=1; i<=N; i++)
    x_n[i] = x[i] - Mu[i];

  for (i=1; i<=N; i++) {
    sum = 0;
    for (j=1; j<=N; j++)
      sum += inv_Sigma[i][j]*x_n[j];
    distance += x_n[i]*sum;
  }

  return distance;
}

