/*---------------------------------------------------------------------------

  Program Name:  pipe_functs.c
  Purpose:       This is the file that contains the main functions for the
                 face recognition software.  The idea of the software is to
		 get an image (passed in), and then take that image, 
		 normalize it, and project it onto an eigenface basis in
		 order to compute a description of the image.  
                  
  Author:        Aaron Thomason
                 David Becker
		 Mark Shackleton

  Date:          Sept. 21, 1994 (pipe_main.c)
                 Nov. 30, 1995 (as pipe_functs.c)

-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

#include "api-defines.h"
#include "api-structs.h"

extern void timestamp(char *label);

int gui_do_correlation = 0;   /* GUI interface do_correlation settings */
int gui_do_correlation2 = 0;

/* structure for qsort */
typedef struct s_dist_entry {
    float   distance;
    int     ix;
} dist_entry_t;

/*---------------------------------------------------------------------------*/

/* global to this module */
static srchparams_t MS_sparams, EIG_sparams, MLD_sparams;
static int MS_nfeatures, EIG_nfeatures, MOD_nfeatures;

static float **ms_variances;
static int   ms_vm, ms_vn;
static float **ms_umap_musigma[MAX_NUM_TEMPLATES];
static int   ms_umap_m, ms_umap_n;
static float **ms_spatial_musigma[MAX_NUM_TEMPLATES];
static int   ms_spatial_m, ms_spatial_n;

static float **eig_variances;
static int   eig_vm, eig_vn;
static float **eig_umap_musigma[MAX_NUM_TEMPLATES];
static int   eig_umap_m, eig_umap_n;
static float **eig_spatial_musigma[MAX_NUM_TEMPLATES];
static int   eig_spatial_m, eig_spatial_n;
static float **windows;
static int   eig_w_m, eig_w_n;

static float ***eigimages;
static int   neigframe = 0;
static int   eigpic_nrow, eigpic_ncol;

static float **mask;

static int   de_flag = 0;
static int   do_residual = 0;
static int   MS_max_N, EIG_max_N;
static float ***MS_template, ***EIG_template;

/*---------------------------------------------------------------------------*/

/***
  This function initialises all the main variables used throughout the
  face location and recognition code. It loads all the required static data
  from data files.
***/

int mit_preload()
{

    FILE  *fp;
    int   MOD_M;
    int   MS_nframe;
    int   MS_sets;
    int   MS_bytes_pixel;
    int   MS_nrow, MS_tnrow, EIG_nrow;
    int   MS_ncol, MS_tncol, EIG_ncol;
    int   i, j, numscales;
    int   geo_nrow, geo_ncol;
    int   *MS_N_dim, *EIG_N_dim; /* the dimensionality of each template */
    int   *MS_M_dim, *EIG_M_dim; /* # of eigenvectors for each template */ 
    int   **MS_templatesize, **EIG_templatesize;
    float *scales;
    float **model;
    float **geo_M;

    char  filename[MAX_CHARS];
    char  line[MAX_CHARS];
    
    char          imagefile[MAX_CHARS], ms_eigdir[MAX_CHARS];
    char          compdir[MAX_CHARS], eigdir[MAX_CHARS];
    
    /* required input flags */
    int ms_flag     = 0;
    int eig_flag    = 0;
    int comp_flag   = 0;
    int errflag     = 0;
    int enroll_flag = 0;
    
    int   c;
    
    char *closest[NUM_OF_NEIGHBORS];
    int   neigframe_select = 0;
    int   neigframe_in = 0;
    
    /* --------------------------MASK stuff-----------------------*/
    
    char mask_file[MAX_CHARS];
    int mrow, mcol;
    
    /* ----------------------end MASK stuff-----------------------*/
    
    timestamp("at start");
    
    strcpy(ms_eigdir, "ms_eigsearch_small");  
    ms_flag = 1;
    
    strcpy(eigdir, "eigsearch_dir"); 
    eig_flag = 1;
    
    strcpy(compdir, "eigen"); 
    comp_flag = 1;
    
    /*             command line error check           */ 
    
#ifdef BT_RELEASE
    
    if (errflag || !ms_flag || !eig_flag || !comp_flag)
	
#else
	
	if (errflag || !ms_flag || !eig_flag || !comp_flag || 
	    (k_flag && de_flag) || (enroll_flag && de_flag) ||
	    (!k_flag && !enroll_flag && !de_flag))
	    
#endif
	    {
		fprintf(stderr, "Inconsistent options set\n");
		return(1);
	    }
    
    
    /* ---Begin pre-loading of data here--- */
    
#ifdef _DEBUG_MALLOC_INC
    fprintf(stderr,"DEBUGGING MALLOC ON\n");
    {
	union dbmalloptarg mval;

	mval.i = M_HANDLE_CORE | M_HANDLE_DUMP;
	dbmallopt(MALLOC_WARN, &mval);

	mval.i = M_HANDLE_ABORT;
	dbmallopt(MALLOC_FATAL, &mval);
	/*
	   mval.str = "malloc.log";
	   dbmallopt(MALLOC_ERRFILE, &mval);
	   */
	mval.i = 1;
	dbmallopt(MALLOC_CKCHAIN, &mval);
    }
#endif
    
    
    /* load eigenfaces into memory */
    
    load_eigen(compdir,neigframe_select,neigframe_in,&neigframe,
	       &eigimages, &eigpic_nrow, &eigpic_ncol);
    
    
    /* ---------------MS_EIG  -> Load eigentemplate data--------------------*/
    
    MS_N_dim = ivector(1, MAX_NUM_TEMPLATES);
    MS_M_dim = ivector(1, MAX_NUM_TEMPLATES);
    
    sprintf(filename, "%s/features.bf", ms_eigdir);
    {
	float **m;
	MS_max_N = 0;
	m = read_BIN(filename, &MS_nfeatures, &MS_ncol);

	MS_templatesize = imatrix(1, MS_nfeatures, 1, 2);

	for (i=1; i<=MS_nfeatures; i++) {
	    for (j=1;j<=MS_ncol; j++)
		MS_templatesize[i][j] = (int) m[i][j];

	    MS_N_dim[i] = MS_templatesize[i][1]*MS_templatesize[i][2];
	    if (MS_N_dim[i]>MS_max_N)
		MS_max_N = MS_N_dim[i];
	}
	free_matrix(m, 1, MS_nfeatures, 1, MS_ncol);
    }
    MS_template = f3tensor(1, MAX_NUM_SEARCH_EIGS,
			   1, MS_max_N, 1, MAX_NUM_TEMPLATES);
    for (i=1; i<= MS_nfeatures; i++) {
	sprintf(filename, "%s/template%d.bf", ms_eigdir, i);
	MS_template[i] = read_BIN(filename,  &MS_tnrow, &MS_tncol);

	MS_M_dim[i] = MS_tnrow;
	if (MS_tncol != MS_N_dim[i])
	    fprintf(stderr, "Template sizes in BF don't match definition file");
    }
    
    
    /* ---------------MS_EIG  -> Load variance data--------------------*/
    
    if (DO_MAHALANOBIS) {

	sprintf(filename,"%s/variances.bf",ms_eigdir);
	ms_variances = read_BIN(filename, &ms_vm, &ms_vn);

	if (ms_vm!=MS_nfeatures) {
	    fprintf(stderr,
		    "ERROR: variances.bf has %d rows! (must have %d)\n\n",
		    ms_vm, MS_nfeatures);
	    return(1);
	}
	if (EIG_LAST>ms_vn) {
	    fprintf(stderr,
		    "ERROR: variances.bf has %d cols (must have atleast %d)\n\n",
		    ms_vn, EIG_LAST);
	    return(1);
	}
    }
    
    
    /*-----------------MS_EIG -> umap_musigma stuff--------------------------*/
    
    if (DO_UNITMAP) {

	float **C, **invC;
	int k;

	C = matrix(1,2,1,2);
	invC = matrix(1,2,1,2);
	for (k=1; k<=MS_nfeatures; k++) {
	    sprintf(filename,"%s/umap_musigma%d.bf",ms_eigdir,k);
	    ms_umap_musigma[k] = read_BIN(filename, &ms_umap_m, &ms_umap_n );
     

	    /* invert the covariance matrix portion of ms_umap_musigma[k] */
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++)
		    C[i][j] = ms_umap_musigma[k][i+1][j];
	    matrix_inverse(C, 2, invC);
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++)
		    ms_umap_musigma[k][i+1][j] = invC[i][j];
	}  
	free_matrix(C,1,2,1,2);
	free_matrix(invC,1,2,1,2);
    }
    
    /*-----------------MS_EIG -> spatial_musigma stuff------------------------*/
    
    if (DO_SPATIAL) {

	float **C, **invC;
	int k;

	C = matrix(1,2,1,2);
	invC = matrix(1,2,1,2);
	for (k=1; k<=MS_nfeatures; k++) {
	    sprintf(filename,"%s/spatial_musigma%d.bf",ms_eigdir,k);
	    ms_spatial_musigma[k] = read_BIN(filename, 
					     &ms_spatial_m, &ms_spatial_n);
     
	    /* invert the covariance matrix portion of umap_musigma[k] */
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++)
		    C[i][j] = ms_spatial_musigma[k][i+1][j];
	    matrix_inverse(C, 2, invC);
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++)
		    ms_spatial_musigma[k][i+1][j] = invC[i][j];
	  
	}  
	free_matrix(C,1,2,1,2);
	free_matrix(invC,1,2,1,2);
    }
    
    /*-----------------MS_EIG -> Load Scalefile------------------------------*/
    
    scales = vector(1, MAX_NUM_SCALES);
    sprintf(filename, "%s/%s", ms_eigdir, SCALEFILE);
    if (!(fp = fopen(filename, "r"))) {
      fprintf(stderr, "ERROR Could not open scale file %s \n\n", filename);
      exit(1);
    }
    i = 1;
    while (fgets(line, MAX_CHARS, fp)) {
      if (strncmp(line, "#", 1) != 0 && strlen(line)>1) {
	sscanf(line, "%f", scales+i);
	i++;
      }
    }
    numscales = i-1;
    fclose(fp);
    if (numscales > MAX_NUM_SCALES) {
      fprintf(stderr,"The number of scales in the scalefile, %s\n",SCALEFILE);
      fprintf(stderr,"is greater than the maximum currently allowed, %d.\n",
	      MAX_NUM_SCALES);
      fprintf(stderr,"Either remove some scales from the scalefile or\n");
      fprintf(stderr,"change MAX_NUM_SCALES in the defines.h file.\n");
      exit(1);
    }

    /*-----------------MS_EIG -> Load Descriptor---------------------------*/
    
    read_descriptor(ms_eigdir, &MS_nframe, &MS_sets, &MS_bytes_pixel,
		    &MS_ncol, &MS_nrow);
    /*
    MS_nrow = MS_EIGSEARCH_NROW;
    MS_ncol = MS_EIGSEARCH_NCOL;
    */

    if (MS_sets>1)
	fprintf(stderr,"Input files must be single-set DAT files!");
    
    /*--------------------EIG -> Load eigentemplate data ------------------*/
    
    EIG_N_dim = ivector(1, MAX_NUM_TEMPLATES);
    EIG_M_dim = ivector(1, MAX_NUM_TEMPLATES);
    
    sprintf(filename, "%s/features.bf", eigdir);
    {
	float **m;
	EIG_max_N = 0;
	m = read_BIN(filename, &EIG_nfeatures, &EIG_ncol);

	EIG_templatesize = imatrix(1, EIG_nfeatures, 1, 2);

	for (i=1; i<=EIG_nfeatures; i++) {
	    for (j=1; j<=EIG_ncol; j++)
		EIG_templatesize[i][j] = (int) m[i][j];
	    EIG_N_dim[i] = EIG_templatesize[i][1]*EIG_templatesize[i][2];
	    if (EIG_N_dim[i]>EIG_max_N)
		EIG_max_N = EIG_N_dim[i];
	}
	EIG_template = f3tensor(1, MAX_NUM_SEARCH_EIGS, 
				1, EIG_max_N, 1, MAX_NUM_TEMPLATES);
	free_matrix(m, 1, EIG_nfeatures, 1, EIG_ncol);
    }
    
    
    for (i=1; i<=EIG_nfeatures; i++) {
	sprintf(filename,"%s/template%d.bf",eigdir,i);
	EIG_template[i] = read_BIN(filename, &EIG_nrow, &EIG_ncol);
	EIG_M_dim[i] = EIG_nrow;
	if (EIG_ncol!=EIG_N_dim[i])
	    myerror("Template sizes in BF don't match those is definition file");
    }
    
    /*--------------------EIG -> variance stuff ---- ------------------*/
    
    if (DO_MAHALANOBIS) {

	sprintf(filename,"%s/variances.bf",eigdir);
	eig_variances = read_BIN(filename, &eig_vm, &eig_vn);

	if (eig_vm!=EIG_nfeatures) {
	    fprintf(stderr,
		    "ERROR: variances.bf has %d rows! (must have %d)\n\n",
		    eig_vm, EIG_nfeatures);
	    return(1);
	}
	if (EIG_LAST2>eig_vn) {
	    fprintf(stderr,
		    "ERROR: variances.bf has %d cols (must have atleast %d)\n\n",
		    eig_vn, EIG_LAST2);
	    return(1);
	}
    }
    
    /*--------------------EIG -> unitmap stuff -----------------------*/
    
    if (DO_UNITMAP) {

	float **C, **invC;
	int k;

	C = matrix(1,2,1,2);
	invC = matrix(1,2,1,2);
	for (k=1; k<=EIG_nfeatures; k++) {
	    sprintf(filename,"%s/umap_musigma%d.bf", eigdir,k);
	    eig_umap_musigma[k] = read_BIN(filename, &eig_umap_m, &eig_umap_n);

     
	    /* invert the covariance matrix portion of umap_musigma[k] */
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++) {
		    C[i][j] = eig_umap_musigma[k][i+1][j];
		}
	    matrix_inverse(C, 2, invC);
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++)
		    eig_umap_musigma[k][i+1][j] = invC[i][j];

	}
	free_matrix(C,1,2,1,2);
	free_matrix(invC,1,2,1,2);
    }
    
    /*--------------------EIG -> spatial stuff -----------------------*/
    
    if (DO_SPATIAL2) {

	float **C, **invC;
	int k;
	C = matrix(1,2,1,2);
	invC = matrix(1,2,1,2);

	for (k=1; k<=EIG_nfeatures; k++) {
	    sprintf(filename,"%s/spatial_musigma%d.bf",eigdir,k);
	    eig_spatial_musigma[k] = read_BIN(filename, 
					      &eig_spatial_m, &eig_spatial_n);

	    /* invert the covariance matrix portion of spatial_musigma[k] */
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++)
		    C[i][j] = eig_spatial_musigma[k][i+1][j];
	    matrix_inverse(C, 2, invC);
	    for (i=1; i<=2; i++)
		for (j=1; j<=2; j++)
		    eig_spatial_musigma[k][i+1][j] = invC[i][j];

	}  
	free_matrix(C,1,2,1,2);
	free_matrix(invC,1,2,1,2);
    }
    
    /*--------------------EIG -> windows stuff -----------------------*/
    
    if (DO_WINDOWS) {

	sprintf(filename,"%s/windows.bf", eigdir);
	windows = read_BIN(filename, &eig_w_m, &eig_w_n);

	if (eig_w_m != EIG_nfeatures || eig_w_n!=4) {
	    fprintf(stderr,"ERROR: windows.bf has wrong dimensions: %d-by-%d \n",
		    eig_w_m, eig_w_n);
	    return(2);
	}
    }
    
    /*----------------------MODEL.BF LOAD----------------------------------*/
    
    sprintf(filename, "%s/model.bf", eigdir);
    model = read_BIN(filename, &MOD_M, &MOD_nfeatures);
    
    
    /*-------------------DETECT GEOMETRY FILE------------------------------*/
    sprintf(filename, "%s/geometry_musigma.bf", eigdir);
    geo_M = read_BIN(filename, &geo_nrow, &geo_ncol);
    
    
    /*--------------------LOAD MASK FILE -----------------------*/
    
    sprintf(mask_file,"%s/%s",eigdir,MASKFILE);
    mask = read_BIN(mask_file, &mrow, &mcol);
    
    if ((mrow != SCALEWARP_NROW) && (mcol != SCALEWARP_NCOL))
	{
	    printf("mask rows / columns do not match descriptor file\n ");
	    return(1);
	}
    
    /* ---End pre-loading here---*/
    
    timestamp("End preload");
    
#ifdef _DEBUG_MALLOC_INC
    printf("MALLOC USAGE AFTER PRE-LOAD = %ld bytes...\n", malloc_inuse(NULL));
    fflush(stdout);
    malloc_dump(2);
#endif
    
    /* Assign the params for srchparams_t to pass to the ms_eigsearch proc */
    
    MS_sparams.nfeatures       = MS_nfeatures;
    MS_sparams.templatesize    = MS_templatesize;
    MS_sparams.template        = MS_template;
    MS_sparams.N_DIM           = MS_N_dim;
    MS_sparams.M_DIM           = MS_M_dim;
    MS_sparams.sets            = MS_sets;
    MS_sparams.bytes_pixel     = MS_bytes_pixel;
/* setup later...
    MS_sparams.ncol            = MS_ncol;
    MS_sparams.nrow            = MS_nrow;
*/
    MS_sparams.geo_nrow        = MS_tnrow;
    MS_sparams.geo_ncol        = MS_tncol;
    MS_sparams.numscales       = numscales;
    MS_sparams.max_N           = MS_max_N;
    MS_sparams.do_graymap      = DO_GRAYMAP;
    MS_sparams.do_unitmap      = DO_UNITMAP;
    MS_sparams.do_mahalanobis  = DO_MAHALANOBIS;
    MS_sparams.multiple_search = MULTIPLE_HEADSEARCH;
    MS_sparams.do_spatial      = DO_SPATIAL;
    MS_sparams.eig_first       = EIG_FIRST;
    MS_sparams.eig_last        = EIG_LAST;
    MS_sparams.scales          = scales;

    /* Assign the params for srchparams_t to pass to eigsearch proc */
    EIG_sparams.nfeatures      = EIG_nfeatures;
    EIG_sparams.templatesize   = EIG_templatesize;
    EIG_sparams.template       = EIG_template;
    EIG_sparams.N_DIM          = EIG_N_dim;
    EIG_sparams.M_DIM          = EIG_M_dim;
    EIG_sparams.sets           = MS_sets;
    EIG_sparams.bytes_pixel    = MS_bytes_pixel;
    EIG_sparams.ncol           = SCALEWARP_NCOL;
    EIG_sparams.nrow           = SCALEWARP_NROW;
    EIG_sparams.geo_nrow       = MS_tnrow;
    EIG_sparams.geo_ncol       = MS_tncol;
    EIG_sparams.max_N          = EIG_max_N;
    EIG_sparams.do_graymap     = DO_GRAYMAP2;
    EIG_sparams.do_unitmap     = DO_UNITMAP2;
    EIG_sparams.do_minima      = DO_MINIMA;
    EIG_sparams.do_mahalanobis = DO_MAHALANOBIS2;
    EIG_sparams.do_spatial     = DO_SPATIAL2;
    EIG_sparams.do_windows     = DO_WINDOWS;
    EIG_sparams.eig_first      = EIG_FIRST2;
    EIG_sparams.eig_last       = EIG_LAST2;

    /* Assign the params for srchparams_t to pass to ml_detect proc */
    MLD_sparams.nfeatures   = MOD_nfeatures;
    MLD_sparams.sets        = MS_sets;
    MLD_sparams.bytes_pixel = MS_bytes_pixel;
    MLD_sparams.ncol        = SCALEWARP_NCOL;
    MLD_sparams.nrow        = SCALEWARP_NROW;
    MLD_sparams.model       = model;
    MLD_sparams.max_N       = MOD_M;
    MLD_sparams.geo_nrow    = geo_nrow;
    MLD_sparams.geo_ncol    = geo_ncol;
    MLD_sparams.geo_M       = geo_M;
    
#ifndef BT_RELEASE
    
    if (video_flag)
	video_init();

#endif

    return(0);

} /* mit_preload */

/*---------------------------------------------------------------------------*/

/***
  This function locates a face in an image, determining head position and
  scale and also feature locations. The head position is returned via the
  pface_image->position structure.

  hier_flag     - Flag: if set, do hierarchical head location.
  *pface_image  - The image to find the head in. Also contains the feature
                  position details which are instantiated here after face
		  location is complete.
***/

int mit_face_locate(int hier_flag, face_image_t *pface_image)
{
    int            i,j;
    int            xres, yres;
    unsigned char  *image;
    float          **orig_im;
    imagparams_t   LOC_imgparams;

    /*
    if (pface_image->format != 0  || pface_image->pixel_aspect != 1.0) {
        fprintf(stderr,"mit_face_locate: Image of wrong format"
		       " or pixel aspect ratio\n");
	return(1);
    }
    */

    if (pface_image->format != 0) {
        fprintf(stderr,"mit_face_locate: Image of wrong format"
		       " - must be raw greyscale\n");

	return(1);
    }

    xres = pface_image->xres;
    yres = pface_image->yres;
    image = pface_image->data;

    orig_im = matrix(1, yres, 1, xres);

    /* copy into float image indexed from (1,1) */
    for (i = 1; i <= yres; i++)
      for (j = 1; j <= xres; j++)
	orig_im[i][j] = (float)(*(image + (i-1)*xres + (j-1)));

    MS_sparams.ncol = xres;
    MS_sparams.nrow = yres;
    MS_sparams.orig_image = orig_im;
    MS_sparams.do_correlation = gui_do_correlation;
    EIG_sparams.do_correlation  = gui_do_correlation2;

#ifndef BT_RELEASE
    
    /* get image from video */
    
    {
	float *temp_im;
	char color_file[MAX_CHARS];
	float **input_im;
	int M = VID_CUT_ROW;
	int N = VID_CUT_COL;
	int M_out = VID_CUT_ROW;
	int N_out = VID_CUT_COL;
	int loops;
	imagparams_t LOC_imgparams;

	temp_im = (float *)malloc(VID_ROW*VID_COL*sizeof(float));
	sprintf(color_file,"%s%s",COLOR_DIR,imagefile);
	get_video_image(temp_im,color_file,de_flag);
	    
	for (i = 1; i <= MS_nrow; i++)
	    for (j = 1; j <= MS_ncol; j++)
		orig_im[i][j] = 0.0;
	    
	input_im = matrix(1,M,1,N);
	    
	/* Pick out appropriate values from digitized image */
	    
	for (i = 1; i <= M; i++)
	    for (j = 1; j <= N; j++)
		input_im[i][j] = temp_im
		    [VID_COL*(i-1)+(j-1+(int)((VID_COL - VID_ROW) / 2))];
	    
	/* ----- decimate image with GP-filter until correct size ----- */
	    
	loops = 0;
	while ((VID_CUT_COL / (pow(2,loops))) > MS_ncol)
	    loops++;
	for (i = 0; i < loops-1; i++)
	    subsample(input_im,M_out,N_out,input_im,&M_out,&N_out,GPFILTER);
	if (loops != 0)
	    subsample(input_im,M_out,N_out,orig_im,&M_out,&N_out,GPFILTER);
	    
	free_matrix(input_im,1,M,1,N);
	free(temp_im);
    }
    
#endif
    
    timestamp("main:calling pipe_locate");
    
    LOC_imgparams = pipe_locate(MS_sparams, ms_variances, ms_umap_musigma,
				ms_spatial_musigma, EIG_sparams, eig_variances,
				eig_umap_musigma, eig_spatial_musigma, windows, 
				MLD_sparams, NULL, NULL, mask, 
				de_flag, hier_flag);

    free_matrix(orig_im, 1, yres, 1, xres);

    /* copy out all the face position details now known */
    /* NOTE: x&y get swapped from MIT back to conventional meaning */

    pface_image->position.head_pos.x = LOC_imgparams.best_head_y;
    pface_image->position.head_pos.y = LOC_imgparams.best_head_x;
    pface_image->position.head_scale = LOC_imgparams.best_scale * SCALEFACTOR;
    pface_image->position.head_status = POS_KNOWN;

    pface_image->position.feature_pos[FACE_LEFT_EYE].x = LOC_imgparams.lefteye_y;
    pface_image->position.feature_pos[FACE_LEFT_EYE].y = LOC_imgparams.lefteye_x;
    pface_image->position.feature_status[FACE_LEFT_EYE] = POS_KNOWN;

    pface_image->position.feature_pos[FACE_RIGHT_EYE].x = LOC_imgparams.righteye_y;
    pface_image->position.feature_pos[FACE_RIGHT_EYE].y = LOC_imgparams.righteye_x;
    pface_image->position.feature_status[FACE_RIGHT_EYE] = POS_KNOWN;

    pface_image->position.feature_pos[FACE_NOSE].x = LOC_imgparams.nose_y;
    pface_image->position.feature_pos[FACE_NOSE].y = LOC_imgparams.nose_x;
    pface_image->position.feature_status[FACE_NOSE] = POS_KNOWN;

    pface_image->position.feature_pos[FACE_MOUTH].x = LOC_imgparams.mouth_y;
    pface_image->position.feature_pos[FACE_MOUTH].y = LOC_imgparams.mouth_x;
    pface_image->position.feature_status[FACE_MOUTH] = POS_KNOWN;

    return(0);

} /* mit_face_locate */

/*---------------------------------------------------------------------------*/

/***
  This routine calculates the coefficients representinf a face's appearance. It
  does this by first normalising the face for position and contrast, masking it,
  then projecting the normalised face onto the set of recognition eigenfaces.

  *pface_image      -  Contains the face image to be processed to provide
                       the coeffs[] describing the face appearance
  *coeffs           -  The array of face coefficients (instantiated here)
  ncoeffs           -  Size of coeffs[] array.
***/

int mit_get_face_coeffs(face_image_t *pface_image, 
			float *coeffs, int ncoeffs)
{
    int            i,j;
    int            xres, yres;
    unsigned char  *image;
    float          **orig_im, **processed_im;
    imagparams_t   MLD_imgparams;
    float          *coeff_vect;

    /*
    if (pface_image->format != 0  || pface_image->pixel_aspect != 1.0) {
        fprintf(stderr,"mit_get_face_coeffs: Image of wrong format"
		       " or pixel aspect ratio\n");
	return(1);
    }
    */

    xres = pface_image->xres;
    yres = pface_image->yres;
    image = pface_image->data;

    if (DO_RIGID) {
        if (pface_image->position.feature_status[DO_RIGID_PT1-1] != POS_KNOWN ||
	    pface_image->position.feature_status[DO_RIGID_PT2-1] != POS_KNOWN) {
	    fprintf(stderr,"mit_get_face_coeffs: the required feature positions"
		           " are not known\n");
	    return(1);
	}
    }
    else {
	/*
        for (i=0; i < NUM_FACIAL_FEATURES; i++) {
	    if (pface_image->position.feature_status[i] != POS_KNOWN) {
	        fprintf(stderr,"mit_get_face_coeffs: the required feature"
			       " positions are not known\n");
		return(1);
	    }
	}
	*/
	fprintf(stderr,"mit_normalise_face: only rigid eye-based face"
		       " registration is implemented!\n");
	return(1);
    }

    orig_im = matrix(1, yres, 1, xres);
    processed_im = matrix(1, SCALEWARP_NROW, 1, SCALEWARP_NCOL);

    /* copy into float image indexed from (1,1) */
    for (i = 1; i <= yres; i++)
      for (j = 1; j <= xres; j++)
	orig_im[i][j] = (float)(*(image + (i-1)*xres + (j-1)));

    /* the input image to be normalised */
    MLD_sparams.nrow = yres;
    MLD_sparams.ncol = xres;
    MLD_sparams.orig_image = orig_im;

    /* initialise details of face feature locations */
    if (DO_RIGID && DO_RIGID_PT1==1 && DO_RIGID_PT2==2) {

        /* note: transpose X&Y back to MIT pipeline coord frame */

	MLD_imgparams.lefteye_x = 
	  pface_image->position.feature_pos[FACE_LEFT_EYE].y;
	MLD_imgparams.lefteye_y = 
	  pface_image->position.feature_pos[FACE_LEFT_EYE].x;

	MLD_imgparams.righteye_x = 
	  pface_image->position.feature_pos[FACE_RIGHT_EYE].y;
	MLD_imgparams.righteye_y = 
	  pface_image->position.feature_pos[FACE_RIGHT_EYE].x;
    }
    else {
	fprintf(stderr,"mit_normalise_face: only rigid eye-based face"
		       " registration is implemented!\n");
	return(1);
    }

    /* normalise the image for feature positions, contrast & apply mask */
    normalise_face(MLD_sparams, MLD_imgparams, mask, processed_im );

    if (ncoeffs > NUM_COEFFS) {
        ncoeffs = NUM_COEFFS;
	fprintf(stderr,"mit_get_face_coeffs: num coeffs reduced\n");
    }
    if (ncoeffs > neigframe) {
        ncoeffs = neigframe;
	fprintf(stderr,"mit_get_face_coeffs: num coeffs reduced (@2)\n");
    }
    
    /* allocate space for coefficient vector and fill it */
    coeff_vect = vector(1, NUM_COEFFS);
    
    timestamp("main:calling make_coeffs");
    
    make_coeffs(SCALEWARP_NROW, SCALEWARP_NCOL, eigpic_nrow, eigpic_ncol,
		do_residual, neigframe, processed_im, coeff_vect, eigimages);

    for (i=0; i < ncoeffs; i++)
        coeffs[i] = coeff_vect[i+1];

    free_matrix(orig_im, 1, yres, 1, xres);
    free_matrix(processed_im, 1, SCALEWARP_NROW, 1, SCALEWARP_NCOL);
    free_vector(coeff_vect,1,NUM_COEFFS);

    return(0);

} /* mit_get_face_coeffs */

/*---------------------------------------------------------------------------*/

static int compare_distance( dist_entry_t *pd1, dist_entry_t *pd2 )
{
    if (pd1->distance < pd2->distance)
	return(-1);
    else if (pd1->distance > pd2->distance)
	return(1);
    else
	return(0);

} /* compare_distance */

/*---------------------------------------------------------------------------*/

int mit_match_face_coeffs(float **known_coeffs, int num_known,
			  float *coeffs, int ncoeffs, 
			  int *match_ix, int *pnum_to_match)
{    
    int            i;
    dist_entry_t   *dist;

    timestamp("mit_match_face_coeffs: doing match");

    if ((dist = calloc(num_known, sizeof(dist_entry_t))) == NULL) {
	fprintf(stderr,"mit_match_face_coeffs: calloc failed!\n");
	return(1);
    }
    
    for (i=0; i < num_known; i++) {
	dist[i].distance = find_distance(coeffs, known_coeffs[i], ncoeffs);
	dist[i].ix = i;
    }

    /* sort the array by ascending distance */
    if (num_known > 1)
	qsort(dist, num_known, sizeof(dist_entry_t), 
	      (int (*)(const void *, const void *))compare_distance);
    if (*pnum_to_match > num_known)
	*pnum_to_match = num_known;

    for (i=0; i < *pnum_to_match; i++)
	match_ix[i] = dist[i].ix;

    free(dist);

    return(0);

} /* mit_match_face_coeffs */

/*---------------------------------------------------------------------------*/

int mit_finish()
{
    int            i;

    timestamp("mit_finish: finishing");
    
    free_matrix(ms_variances,1,ms_vm,1,ms_vn);
    free_matrix(eig_variances,1,eig_vm,1,eig_vn);
    for (i=1; i<=MS_nfeatures; i++) {
	free_matrix(ms_umap_musigma[i], 1, ms_umap_m, 1, ms_umap_n);
	free_matrix(ms_spatial_musigma[i],1,ms_spatial_m,1,ms_spatial_n);
    }
    for (i=1; i<=EIG_nfeatures; i++) {
	free_matrix(eig_umap_musigma[i], 1, eig_umap_m, 1, eig_umap_n);
	free_matrix(eig_spatial_musigma[i],1,eig_spatial_m,1,eig_spatial_n);
    }
    free_matrix(windows,1,eig_w_m,1,eig_w_n);
    free_matrix(mask,1,SCALEWARP_NROW,1,SCALEWARP_NCOL);

    free_f3tensor(MS_template, 1, MAX_NUM_SEARCH_EIGS, 
		  1, MS_max_N, 1, MAX_NUM_TEMPLATES);
    free_f3tensor(EIG_template, 1, MAX_NUM_SEARCH_EIGS, 
		  1, EIG_max_N, 1, MAX_NUM_TEMPLATES);
    
    return(0);
    
} /* mit_finish */

/*---------------------------------------------------------------------------*/
