/*---------------------------------------------------------------------------

  Program Name:  pipe_main.c
  Purpose:       This is the file that contains the main for the face
                 recognition software.  The idea of the software is to
		 get an image, either via a video feed or by reading
		 it off disk, and then take that image, normalize it,
		 and project it onto an eigenface basis in order to
		 compute a description of the image.  
                  
  Author:        Aaron Thomason
                 David Becker

  Date:          Sept. 21, 1994

-----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

extern void timestamp(char *label);

/* ----------------------Command-line Parsing Stuff--------------------------*/

#ifdef BT_RELEASE

#define OPTIONS "s:e:c:k:rnl:h"

#else

#define OPTIONS "s:e:c:k:rnl:vdh"

#endif

extern int  optind;
extern char *optarg;

/*-----------------------------MAIN----------------------------------------*/

int main(int argc, char *argv[])
{

  char        comline[MAX_CHARS];    /* used to store the command line */
  char        progname[MAX_CHARS];   /* used to store program name */

#ifdef BT_RELEASE

  char        *usage = 
    "\t-s ms_eigdir -e eigdir -c eigendir [-k knownfacesdir]\n"
    "\t[-r residual] [-n enroll_flag] [-l list] [-h hierarchy]\n";

#else

  char        *usage = 
    "\t-s ms_eigdir -e eigdir -c eigendir [-k knownfacesdir]\n"
    "\t[-r residual] [-n enroll_flag] [-l list] [-v video]\n"
    "\t[-d add to de photobook] [-h hierarchy]\n";

#endif
  
  char *help = "Pipe: \n\n"
    "-s ms_eigdir   \t ms_eigsearch .bf directory\n"
    "-e eigdir      \t eigsearch .bf directory\n"
    "-c eigendir    \t directory with eigenfaces\n"
    "-k knowndir    \t directory with coeffs to compare image to or\n" 
    "               \t directory to put coefficients in\n"
    "               \t only fill this in with -n or to identify from\n"
    "               \t the knowndir, but not with -d\n"
    "-r residual    \t flag for whether or not to compute residuals\n"
    "-n enroll      \t include to enroll image, leave off to get image's\n"
    "               \t neigbors\n"
    "-l list        \t will read files to process from given list\n"

#ifndef BT_RELEASE

    "-v video       \t include this flag to digitize from video\n"
    "               \t NOTE: You can't have -l and -v\n"
    "-d database    \t add images to Digital Expressions database\n"

#endif

    "-h hierarchy   \t include flag to do hierarchical headsearch\n";

  /* Command line defaults */

  int do_residual = 0;     /*  default is to not do residual display */
  
  /* Search variables      */

  int   MS_max_N, EIG_max_N, MOD_M;
  int   MS_nfeatures, EIG_nfeatures, MOD_nfeatures;
  int   MS_nframe;
  int   MS_sets;
  int   MS_bytes_pixel;
  int   MS_nrow, MS_tnrow, EIG_nrow;
  int   MS_ncol, MS_tncol, EIG_ncol;
  int   i, j, numscales, nrow, ncol;
  int   geo_nrow, geo_ncol;
  int   *MS_N_dim, *EIG_N_dim; /* the dimensionality of each template */
  int   *MS_M_dim, *EIG_M_dim; /* # of eigenvectors for each template */ 
  int   **MS_templatesize, **EIG_templatesize;
  float *scales;
  float **model;
  float **geo_M;
  float ***MS_template, ***EIG_template;
  char  filename[MAX_CHARS];
  char  line[MAX_CHARS];

  FILE  *fp, *list_file_ptr;

  srchparams_t  MS_sparams, EIG_sparams, MLD_sparams;
  char          imagefile[MAX_CHARS], ms_eigdir[MAX_CHARS];
  char          compdir[MAX_CHARS], eigdir[MAX_CHARS];
  char          k_dir[MAX_CHARS], list_file[MAX_CHARS];
  image_t       orig_im, processed_im;
  float         *coeff_vect;
  float         ***eigimages;
  
  /* required input flags */
  int ms_flag     = 0;
  int eig_flag    = 0;
  int comp_flag   = 0;
  int errflag     = 0;
  int k_flag      = 0;
  int enroll_flag = 0;
  int list_flag   = 0;
  int de_flag     = 0;
  int hier_flag   = 0;

  int   c;
  int   list_counter = 0;
  int   num_in_list = 0;

  int   num_k_coeffs=0;
  int   num_k=0;
  char *closest[NUM_OF_NEIGHBORS];
  int   neigframe_select = 0;
  int   neigframe_in = 0;
  int   neigframe = 0;

  Kface *kfaces;    /* typedef for information of known faces */


  /* ---------------------MS_EIGSEARCH stuff--------------------*/

  float **ms_variances;
  float **ms_umap_musigma[MAX_NUM_TEMPLATES];
  float **ms_spatial_musigma[MAX_NUM_TEMPLATES];
  int     ms_vm, ms_vn, ms_umap_m, ms_umap_n, ms_spatial_m, ms_spatial_n;

  /* -----------------end MS_EIGSEARCH stuff--------------------*/

  /* ---------------------EIGSEARCH stuff-----------------------*/

  float **eig_variances;
  float **eig_umap_musigma[MAX_NUM_TEMPLATES];
  float **eig_spatial_musigma[MAX_NUM_TEMPLATES];
  float **w;
  int     eig_vm, eig_vn, eig_umap_m, eig_umap_n, eig_spatial_m;
  int     eig_spatial_n, eig_w_m, eig_w_n;

  /* -----------------end EIGSEARCH stuff-----------------------*/

  /* --------------------------MASK stuff-----------------------*/

  char mask_file[MAX_CHARS];
  float **mask;
  int mrow, mcol;

  /* ----------------------end MASK stuff-----------------------*/

timestamp("at start");

  /* setup command line string */

  sprintf(progname,"%s",argv[0]);
  for (i=0; i<argc; i++)
    strcat(comline, argv[i]), strcat(comline, " ");

  /* Command line parse */

  while ((c = getopt(argc, argv, OPTIONS)) != EOF)
    switch(c) {

    case 's':
      strcpy(ms_eigdir, optarg);
      ms_flag = 1;
      break;
    case 'e':
      strcpy(eigdir, optarg);
      eig_flag = 1;
      break;
    case 'c':
      strcpy(compdir, optarg);
      comp_flag = 1;
      break;
    case 'k':
      strcpy(k_dir, optarg);
      k_flag = 1;
      break;
    case 'r':
      do_residual = 1;
      break;
    case 'n':
      enroll_flag = 1;
      break;
    case 'l':
      list_flag = 1;
      strcpy(list_file, optarg);
      break;

#ifndef BT_RELEASE

    case 'v':
      video_flag = 1;
      break;
    case 'd':
      de_flag = 1;
      break;

#endif

    case 'h':
      hier_flag = 1;
      break;
    case '?':
      errflag = 0;
      break;
      
    }

  /*             command line error check           */ 

#ifdef BT_RELEASE

  if (errflag || !ms_flag || !eig_flag || !comp_flag || 
     (!k_flag && !enroll_flag))

#else

  if (errflag || !ms_flag || !eig_flag || !comp_flag || 
      (video_flag && list_flag) || (k_flag && de_flag) ||
      (enroll_flag && de_flag) ||
      (!k_flag && !enroll_flag && !de_flag))

#endif
    {
      fprintf(stderr, "\nUSAGE:\n%s %s\n%s\n", progname, usage, help);
      exit(1);
    }


/* ---Begin pre-loading of data here--- */

#ifdef _DEBUG_MALLOC_INC
  fprintf(stderr,"DEBUGGING MALLOC ON\n");
  {
    union dbmalloptarg mval;

    mval.i = M_HANDLE_CORE | M_HANDLE_DUMP;
    dbmallopt(MALLOC_WARN, &mval);

    mval.i = M_HANDLE_ABORT;
    dbmallopt(MALLOC_FATAL, &mval);
/*
    mval.str = "malloc.log";
    dbmallopt(MALLOC_ERRFILE, &mval);
*/
    mval.i = 1;
    dbmallopt(MALLOC_CKCHAIN, &mval);
  }
#endif


  /* if given a list, load filenames here */

  if (list_flag)
    {
      if ((list_file_ptr = fopen(list_file,"r")) == NULL)
	{
	  fprintf(stderr,"Error opening list file %s\n",list_file);
	  exit(2);
	}
      while (fscanf(list_file_ptr,"%s",imagefile) != EOF)
	{
	  if ((imagefile[0] != '#') && (strlen(imagefile) > 1))
	    num_in_list++;
	}
      rewind(list_file_ptr);
    }


  /* load known faces if not enrolling this face */

  if (k_flag && !enroll_flag)
    load_known(&kfaces,&num_k_coeffs,&num_k,k_dir);

  
  /* load eigenfaces into memory */
  
  load_eigen(compdir,neigframe_select,neigframe_in,&neigframe,
	     &eigimages, &nrow, &ncol);

  
  /* ---------------MS_EIG  -> Load eigentemplate data--------------------*/

  MS_N_dim = ivector(1, MAX_NUM_TEMPLATES);
  MS_M_dim = ivector(1, MAX_NUM_TEMPLATES);

  sprintf(filename, "%s/features.bf", ms_eigdir);
  {
    float **m;
    MS_max_N = 0;
    m = read_BIN(filename, &MS_nfeatures, &MS_ncol);

    MS_templatesize = imatrix(1, MS_nfeatures, 1, 2);

    for (i=1; i<=MS_nfeatures; i++) {
      for (j=1;j<=MS_ncol; j++)
        MS_templatesize[i][j] = (int) m[i][j];
      MS_N_dim[i] = MS_templatesize[i][1]*MS_templatesize[i][2];
      if (MS_N_dim[i]>MS_max_N)
        MS_max_N = MS_N_dim[i];
    }
    free_matrix(m, 1, MS_nfeatures, 1, MS_ncol);
  }
  MS_template = f3tensor(1, 51, 1, 475, 1, MAX_NUM_TEMPLATES);
  for (i=1; i<= MS_nfeatures; i++) {
    sprintf(filename, "%s/template%d.bf", ms_eigdir, i);
    MS_template[i] = read_BIN(filename,  &MS_tnrow, &MS_tncol);

    MS_M_dim[i] = MS_tnrow;
    if (MS_tncol != MS_N_dim[i])
      fprintf(stderr, "Template sizes in BF don't match definition file");
  }


  /* ---------------MS_EIG  -> Load variance data--------------------*/

  if (DO_MAHALANOBIS)
    {
      sprintf(filename,"%s/variances.bf",ms_eigdir);
      ms_variances = read_BIN(filename, &ms_vm, &ms_vn);

      if (ms_vm!=MS_nfeatures) 
	{
	  fprintf(stderr,
		  "ERROR: variances.bf has %d rows! (must have %d)\n\n",
		  ms_vm, MS_nfeatures);
	  exit(1);
	}
      if (EIG_LAST>ms_vn) 
	{
	  fprintf(stderr,
		  "ERROR: variances.bf has %d cols (must have atleast %d)\n\n",
		  ms_vn, EIG_LAST);
	  exit(1);
	}
    }
    
     
  /*-----------------MS_EIG -> umap_musigma stuff--------------------------*/

  if (DO_UNITMAP) 
    {
      float **C, **invC;
      int k;

      C = matrix(1,2,1,2);
      invC = matrix(1,2,1,2);
      for (k=1; k<=MS_nfeatures; k++) 
	{
	  sprintf(filename,"%s/umap_musigma%d.bf",ms_eigdir,k);
	  ms_umap_musigma[k] = read_BIN(filename, &ms_umap_m, &ms_umap_n );
     

	  /* invert the covariance matrix portion of ms_umap_musigma[k] */
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++)
	      C[i][j] = ms_umap_musigma[k][i+1][j];
	  matrix_inverse(C, 2, invC);
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++)
	      ms_umap_musigma[k][i+1][j] = invC[i][j];
	}  
      free_matrix(C,1,2,1,2);
      free_matrix(invC,1,2,1,2);
    }

  /*-----------------MS_EIG -> spatial_musigma stuff------------------------*/

  if (DO_SPATIAL) 
    {
      float **C, **invC;
      int k;

      C = matrix(1,2,1,2);
      invC = matrix(1,2,1,2);
      for (k=1; k<=MS_nfeatures; k++) 
	{
	  sprintf(filename,"%s/spatial_musigma%d.bf",ms_eigdir,k);
	  ms_spatial_musigma[k] = read_BIN(filename, 
					   &ms_spatial_m, &ms_spatial_n);
     
	  /* invert the covariance matrix portion of umap_musigma[k] */
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++)
	      C[i][j] = ms_spatial_musigma[k][i+1][j];
	  matrix_inverse(C, 2, invC);
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++)
	      ms_spatial_musigma[k][i+1][j] = invC[i][j];
	  
	}  
      free_matrix(C,1,2,1,2);
      free_matrix(invC,1,2,1,2);
    }


  /*-----------------MS_EIG -> Load Scalefile------------------------------*/

  scales = vector(1, MAX_NUM_SCALES);
  sprintf(filename, "%s/%s", ms_eigdir, SCALEFILE);
  if (!(fp = fopen(filename, "r"))) {
    fprintf(stderr, "ERROR Could not open scale file %s \n\n", filename);
    exit(1);
  }
  i = 1;
  while (fgets(line, MAX_CHARS, fp)) {
    if (strncmp(line, "#", 1) != 0 && strlen(line)>1) {
      sscanf(line, "%f", scales+i);
      i++;
    }
  }
  numscales = i-1;
  fclose(fp);
  if (numscales > MAX_NUM_SCALES)
    {
      fprintf(stderr,"The number of scales in the scalefile, %s\n",SCALEFILE);
      fprintf(stderr,"is greater than the maximum currently allowed, %d.\n",
	      MAX_NUM_SCALES);
      fprintf(stderr,"Either remove some scales from the scalefile or\n");
      fprintf(stderr,"change MAX_NUM_SCALES in the defines.h file.\n");
      exit(1);
    }

  /*-----------------MS_EIG -> Load Descriptor---------------------------*/

  read_descriptor(ms_eigdir, &MS_nframe, &MS_sets, &MS_bytes_pixel,
                  &MS_ncol, &MS_nrow);

  MS_nrow = MS_EIGSEARCH_NROW;
  MS_ncol = MS_EIGSEARCH_NCOL;
  if (MS_sets>1)
    fprintf(stderr,"Input files must be single-set DAT files!");

  /*--------------------EIG -> Load eigentemplate data ------------------*/

  EIG_N_dim = ivector(1, MAX_NUM_TEMPLATES);
  EIG_M_dim = ivector(1, MAX_NUM_TEMPLATES);

  sprintf(filename, "%s/features.bf", eigdir);
  {
    float **m;
    EIG_max_N = 0;
    m = read_BIN(filename, &EIG_nfeatures, &EIG_ncol);

    EIG_templatesize = imatrix(1, EIG_nfeatures, 1, 2);

    for (i=1; i<=EIG_nfeatures; i++) {
      for (j=1; j<=EIG_ncol; j++)
        EIG_templatesize[i][j] = (int) m[i][j];
      EIG_N_dim[i] = EIG_templatesize[i][1]*EIG_templatesize[i][2];
      if (EIG_N_dim[i]>EIG_max_N)
        EIG_max_N = EIG_N_dim[i];
    }
    free_matrix(m, 1, EIG_nfeatures, 1, EIG_ncol);

/* MAS DEBUG
for (i=1; i<=EIG_nfeatures; i++) 
 printf("EIG template size [%d]: r=%d c=%d\n",
  i,EIG_templatesize[i][1],EIG_templatesize[i][2]);
*/

  }

  EIG_template = f3tensor(1, 51, 1, 625, 1, MAX_NUM_TEMPLATES);
  for (i=1; i<=EIG_nfeatures; i++) {
    sprintf(filename,"%s/template%d.bf",eigdir,i);
    EIG_template[i] = read_BIN(filename, &EIG_nrow, &EIG_ncol);

    EIG_M_dim[i] = EIG_nrow;
    if (EIG_ncol!=EIG_N_dim[i])
      myerror("Template sizes in BF don't match those is definition file");
  }

  /*--------------------EIG -> variance stuff ---- ------------------*/

  if (DO_MAHALANOBIS) 
    {
      sprintf(filename,"%s/variances.bf",eigdir);
      eig_variances = read_BIN(filename, &eig_vm, &eig_vn);

      if (eig_vm!=EIG_nfeatures) 
	{
	  fprintf(stderr,
		  "ERROR: variances.bf has %d rows! (must have %d)\n\n",
		  eig_vm, EIG_nfeatures);
	  exit(1);
	}
      if (EIG_LAST2>eig_vn) 
	{
	  fprintf(stderr,
		  "ERROR: variances.bf has %d cols (must have atleast %d)\n\n",
		  eig_vn, EIG_LAST2);
	  exit(1);
	}
    }
    
  /*--------------------EIG -> unitmap stuff -----------------------*/

  if (DO_UNITMAP) 
    {
      float **C, **invC;
      int k;

      C = matrix(1,2,1,2);
      invC = matrix(1,2,1,2);
      for (k=1; k<=EIG_nfeatures; k++) 
	{
	  sprintf(filename,"%s/umap_musigma%d.bf", eigdir,k);
	  eig_umap_musigma[k] = read_BIN(filename, &eig_umap_m, &eig_umap_n);

     
	  /* invert the covariance matrix portion of umap_musigma[k] */
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++) {
	      C[i][j] = eig_umap_musigma[k][i+1][j];
	    }
	  matrix_inverse(C, 2, invC);
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++)
	      eig_umap_musigma[k][i+1][j] = invC[i][j];

	}
      free_matrix(C,1,2,1,2);
      free_matrix(invC,1,2,1,2);
    }

  /*--------------------EIG -> spatial stuff -----------------------*/
   
  if (DO_SPATIAL) 
    {
      float **C, **invC;
      int k;
      C = matrix(1,2,1,2);
      invC = matrix(1,2,1,2);

      for (k=1; k<=EIG_nfeatures; k++) 
	{
	  sprintf(filename,"%s/spatial_musigma%d.bf",eigdir,k);
	  eig_spatial_musigma[k] = read_BIN(filename, 
					    &eig_spatial_m, &eig_spatial_n);

	  /* invert the covariance matrix portion of spatial_musigma[k] */
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++)
	      C[i][j] = eig_spatial_musigma[k][i+1][j];
	  matrix_inverse(C, 2, invC);
	  for (i=1; i<=2; i++)
	    for (j=1; j<=2; j++)
	      eig_spatial_musigma[k][i+1][j] = invC[i][j];

	}  
      free_matrix(C,1,2,1,2);
      free_matrix(invC,1,2,1,2);
    }

  /*--------------------EIG -> windows stuff -----------------------*/

  if (DO_WINDOWS) 
    {
      sprintf(filename,"%s/windows.bf", eigdir);
      w = read_BIN(filename, &eig_w_m, &eig_w_n);

      if (eig_w_m != EIG_nfeatures || eig_w_n!=4) 
	{
	  fprintf(stderr,"ERROR: windows.bf has wrong dimensions: %d-by-%d \n",
		  eig_w_m, eig_w_n);
	  exit(2);
	}

/* MAS DEBUG
for (i=1; i<=EIG_nfeatures; i++) 
 printf("EIG windows size [%d]: rmin=%f rmax=%f cmin=%f cmax=%f\n",
  i,w[i][1],w[i][2],w[i][3],w[i][4]);
*/
    }


  /*----------------------MODEL.BF LOAD----------------------------------*/

  sprintf(filename, "%s/model.bf", eigdir);
  model = read_BIN(filename, &MOD_M, &MOD_nfeatures);

  /*-------------------DETECT GEOMETRY FILE------------------------------*/
  sprintf(filename, "%s/geometry_musigma.bf", eigdir);
  geo_M = read_BIN(filename, &geo_nrow, &geo_ncol);


  /*--------------------LOAD MASK FILE -----------------------*/

  /* ---End pre-loading here---*/

timestamp("End preload");

  sprintf(mask_file,"%s/%s",eigdir,MASKFILE);
  mask = read_BIN(mask_file, &mrow, &mcol);

  
  if ((mrow != SCALEWARP_NROW) && (mcol != SCALEWARP_NCOL))
    {
      printf("mask rows / columns do not match descriptor file\n ");
      exit(1);
    }

#ifdef _DEBUG_MALLOC_INC
  printf("MALLOC USAGE AFTER PRE-LOAD = %ld bytes...\n", malloc_inuse(NULL));
  fflush(stdout);
  malloc_dump(2);
#endif

  /* Assign the params for srchparams_t to pass to the ms_eigsearch proc */

  MS_sparams.nfeatures       = MS_nfeatures;
  MS_sparams.templatesize    = MS_templatesize;
  MS_sparams.template        = MS_template;
  MS_sparams.N_DIM           = MS_N_dim;
  MS_sparams.M_DIM           = MS_M_dim;
  MS_sparams.sets            = MS_sets;
  MS_sparams.bytes_pixel     = MS_bytes_pixel;
  MS_sparams.ncol            = MS_ncol;
  MS_sparams.nrow            = MS_nrow;
  MS_sparams.geo_nrow        = MS_tnrow;
  MS_sparams.geo_ncol        = MS_tncol;
  MS_sparams.numscales       = numscales;
  MS_sparams.max_N           = MS_max_N;
  MS_sparams.do_graymap      = DO_GRAYMAP;
  MS_sparams.do_unitmap      = DO_UNITMAP;
  MS_sparams.do_mahalanobis  = DO_MAHALANOBIS;
  MS_sparams.multiple_search = MULTIPLE_HEADSEARCH;
  MS_sparams.do_spatial      = DO_SPATIAL;
  MS_sparams.eig_first       = EIG_FIRST;
  MS_sparams.eig_last        = EIG_LAST;
  MS_sparams.scales          = scales;

  /* Assign the params for srchparams_t to pass to eigsearch proc */
  EIG_sparams.nfeatures      = EIG_nfeatures;
  EIG_sparams.templatesize   = EIG_templatesize;
  EIG_sparams.template       = EIG_template;
  EIG_sparams.N_DIM          = EIG_N_dim;
  EIG_sparams.M_DIM          = EIG_M_dim;
  EIG_sparams.sets           = MS_sets;
  EIG_sparams.bytes_pixel    = MS_bytes_pixel;
  EIG_sparams.ncol           = SCALEWARP_NCOL;
  EIG_sparams.nrow           = SCALEWARP_NROW;
  EIG_sparams.geo_nrow       = MS_tnrow;
  EIG_sparams.geo_ncol       = MS_tncol;
  EIG_sparams.max_N          = EIG_max_N;
  EIG_sparams.do_graymap     = DO_GRAYMAP2;
  EIG_sparams.do_unitmap     = DO_UNITMAP2;
  EIG_sparams.do_minima      = DO_MINIMA;
  EIG_sparams.do_mahalanobis = DO_MAHALANOBIS2;
  EIG_sparams.do_spatial     = DO_SPATIAL2;
  EIG_sparams.do_windows     = DO_WINDOWS;
  EIG_sparams.eig_first      = EIG_FIRST2;
  EIG_sparams.eig_last       = EIG_LAST2;

  /* Assign the params for srchparams_t to pass to ml_detect proc */
  MLD_sparams.nfeatures   = MOD_nfeatures;
  MLD_sparams.sets        = MS_sets;
  MLD_sparams.bytes_pixel = MS_bytes_pixel;
  MLD_sparams.ncol        = SCALEWARP_NCOL;
  MLD_sparams.nrow        = SCALEWARP_NROW;
  MLD_sparams.model       = model;
  MLD_sparams.max_N       = MOD_M;
  MLD_sparams.geo_nrow    = geo_nrow;
  MLD_sparams.geo_ncol    = geo_ncol;
  MLD_sparams.geo_M       = geo_M;

#ifndef BT_RELEASE

  if (video_flag)
    video_init();

#endif

  /* if not reading from list, loop should go until ctrl-c, so make */
  /* sure that list_counter < num_in_list */

  if (!list_flag)
     list_counter = -1;

  while (list_counter < num_in_list)
    {

timestamp("main:list_counter");
      
      /*----------------original image file----------------------------------*/

      /* if reading from list, get next image */

      if (list_flag)
	{
	  fscanf(list_file_ptr,"%s",imagefile);
	  fprintf(stdout,"Processing %s\n",imagefile);
	  list_counter++;
	}

      /* otherwise get filename from keyboard */

      else
	{
	  
	  fprintf(stdout,
		  "\nReady to process image.  Please enter image filename "
		  "here: ");
	  scanf("%s",imagefile);
	}

      orig_im = matrix(1, MS_nrow, 1, MS_ncol);

      /* If reading file from disk, do this */

#ifndef BT_RELEASE

      if (!video_flag)

#endif

	read_RAW(imagefile,orig_im,MS_EIGSEARCH_NROW,MS_EIGSEARCH_NCOL);

#ifndef BT_RELEASE

      /* Otherwise get image from video */

      else
	{
	  {
	    float *temp_im;
	    char color_file[MAX_CHARS];
	    image_t input_im;
	    int M = VID_CUT_ROW;
	    int N = VID_CUT_COL;
	    int M_out = VID_CUT_ROW;
	    int N_out = VID_CUT_COL;
	    int loops;

	    temp_im = (float *)malloc(VID_ROW*VID_COL*sizeof(float));
	    sprintf(color_file,"%s%s",COLOR_DIR,imagefile);
	    get_video_image(temp_im,color_file,de_flag);

	    for (i = 1; i <= MS_nrow; i++)
	      for (j = 1; j <= MS_ncol; j++)
		orig_im[i][j] = 0.0;

	    input_im = matrix(1,M,1,N);
	    
	    /* Pick out appropriate values from digitized image */

	    for (i = 1; i <= M; i++)
	      for (j = 1; j <= N; j++)
		input_im[i][j] = temp_im
		  [VID_COL*(i-1)+(j-1+(int)((VID_COL - VID_ROW) / 2))];

	    /* ----- decimate image with GP-filter until correct size ----- */

	    loops = 0;
	    while ((VID_CUT_COL / (pow(2,loops))) > MS_ncol)
	      loops++;
	    for (i = 0; i < loops-1; i++)
	      subsample(input_im,M_out,N_out,input_im,&M_out,&N_out,GPFILTER);
	    if (loops != 0)
	      subsample(input_im,M_out,N_out,orig_im,&M_out,&N_out,GPFILTER);
	    
	    free_matrix(input_im,1,M,1,N);
	    free(temp_im);
	  }
	}

      /* if user asked to put in de database, do so here and where */
      /* de_flag is in if statements below */
      
      if (de_flag)
	{
	  sprintf(original_file,"%s%s",PHOTOBOOK_ORIG,imagefile);
	  write_float_as_uchar(orig_im,MS_EIGSEARCH_NROW,MS_EIGSEARCH_NCOL,
			       original_file);
	}

#endif

      MS_sparams.orig_image = orig_im;
      processed_im = matrix(1,SCALEWARP_NCOL,1,SCALEWARP_NROW);

timestamp("main:calling pipe");

      pipe(MS_sparams, ms_variances, ms_umap_musigma,
	   ms_spatial_musigma, EIG_sparams, eig_variances,
	   eig_umap_musigma, eig_spatial_musigma, w, MLD_sparams, 
	   processed_im, imagefile, mask, de_flag, hier_flag);

#ifndef BT_RELEASE

      if (de_flag)
	{
	  sprintf(norm_file,"%s%s",PHOTOBOOK_NORM,imagefile);
	  write_float_as_uchar(processed_im,SCALEWARP_NROW,SCALEWARP_NCOL,
			       norm_file);
	}

#endif	

      /* allocate space for coefficient vector and fill it */
      coeff_vect = vector(1, NUM_COEFFS);

timestamp("main:calling make_coeffs");

      make_coeffs(SCALEWARP_NROW, SCALEWARP_NCOL, nrow, ncol,
		  do_residual, neigframe, processed_im, coeff_vect,eigimages);

      /* Either enroll the image or check it against known faces */

#ifndef BT_RELEASE

      if (de_flag)
	{
	  sprintf(coeff_file,"%s%s",PHOTOBOOK_COEFF,imagefile);
	  WriteBinary(coeff_vect,NUM_COEFFS,coeff_file);
	}
      else 
	{

#endif

timestamp("main:doing enrol/neighbor");

	  if (enroll_flag)
	    enroll(imagefile,k_dir,coeff_vect,neigframe);
	  else
	    {
	      neighbor(kfaces,num_k_coeffs,num_k,coeff_vect,closest);

	      /* Display on screen the named closest people */
	      
	      for (i = 0; i < NUM_OF_NEIGHBORS; i++)
		if (strlen(closest[i]) > 0)
		  fprintf(stdout, "neighbor %d, %s\n", i+1, closest[i]);
	    }

#ifndef BT_RELEASE

	}

#endif

      free_matrix(orig_im, 1, MS_nrow, 1, MS_ncol);
      free_matrix(processed_im, 1, SCALEWARP_NCOL, 1, SCALEWARP_NROW);
      free_vector(coeff_vect,1,NUM_COEFFS);
}

timestamp("main:finished");

  fclose(list_file_ptr);
  free_matrix(ms_variances,1,ms_vm,1,ms_vn);
  free_matrix(eig_variances,1,eig_vm,1,eig_vn);
  for (i=1; i<=MS_nfeatures; i++)
    {
      free_matrix(ms_umap_musigma[i], 1, ms_umap_m, 1, ms_umap_n);
      free_matrix(ms_spatial_musigma[i],1,ms_spatial_m,1,ms_spatial_n);
    }
  for (i=1; i<=EIG_nfeatures; i++)
    {
      free_matrix(eig_umap_musigma[i], 1, eig_umap_m, 1, eig_umap_n);
      free_matrix(eig_spatial_musigma[i],1,eig_spatial_m,1,eig_spatial_n);
    }
  free_matrix(w,1,eig_w_m,1,eig_w_n);
  free_matrix(mask,1,SCALEWARP_NROW,1,SCALEWARP_NCOL);

  return(0);
}









