/*___________________________________________________________________________
  pipeline.c 
  Function pipe takes in an image and search structures and returns an image
  modelwarped (with ms_eigsearch, eigsearch, and ml_detect outputs), masked,
  and contrastmapped.

  Created by Aaron Thomason August 29, 1994

  ---------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/* Mark, this is the program where most of the work gets done.  It */
 /* calls a series of functions which do headfinding, scaling, etc. */
 /* You can print out the warped, masked, etc. image at any point by */
 /* making a call to */
 /* */
 /*               write_RAW_float(filename,image,ncol,nrow) */
 /* */
 /* where the filename what you want to name it, image is a **float */
 /* that is the image you want to write to disk, ncol and nrow are 128 */
 /* in these cases, or whatever SCALEWARP_NROW and SCALEWARP_NCOL are */
 /* defined as in defines.h.*/
 /* Alternatively, if you want your output to be uchars, not floats, */
 /* you should use the function, write_float_as_uchar instead of */
 /* write_RAW_float... */
/*-------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include "defines.h"
#include "util.h"
#include "io.h"
#include "matrix.h"
#include "affine.h"
#include "convolve.h"
#include "functions.h"
#include "faces.h"

extern void timestamp(char *label);

void
  pipe(srchparams_t MS_sparams, 
       float **ms_variances, float **ms_umap_musigma[MAX_NUM_TEMPLATES],
       float **ms_spatial_musigma[MAX_NUM_TEMPLATES],
       srchparams_t EIG_sparams, float **eig_variances, 
       float **eig_umap_musigma[MAX_NUM_TEMPLATES], 
       float **eig_spatial_musigma[MAX_NUM_TEMPLATES], float **w,
       srchparams_t MLD_sparams, image_t image_out,
       char *filename, float **mask, int de_flag, int hier_flag)
{
  imagparams_t MS_imgparams, EIG_imgparams, MLD_imgparams;
  float        **ERR_minima;
  int          **X_minima, **Y_minima;
  int          *N_minima;
  char         head_file[MAX_CHARS];
  image_t      scale_im;
  int          num_minima;

  /* head finding */

timestamp("pipe:ms_search:1");

  MS_imgparams = ms_search(MS_sparams, EIG_sparams, MLD_sparams, 
			   &MLD_imgparams, hier_flag, ms_variances, 
			   ms_umap_musigma, ms_spatial_musigma, eig_variances,
			   eig_umap_musigma, eig_spatial_musigma, w);

timestamp("pipe:scale_warp:1");

  scale_im = matrix(1,SCALEWARP_NROW,1,SCALEWARP_NCOL);
  scale_warp(MS_sparams, MS_imgparams.best_head_x, MS_imgparams.best_head_y,
	    MS_imgparams.best_scale, scale_im);

#ifndef HP_RELEASE

  /* if de_flag then put into de database */

  if (de_flag)
    {
      sprintf(head_file,"%s%s",PHOTOBOOK_HEAD,filename);
      write_float_as_uchar(scale_im,SCALEWARP_NROW,SCALEWARP_NCOL,head_file);
    }

#endif

  /* feature looking (picks out possible targets.  If doing */
  /* hierarchical head detection, you already have have the locations */
  /* for the features from that, so no need to do it again. */

  if (!hier_flag)
    {

timestamp("pipe:eig_search:1 (not hierarchical)");

      EIG_sparams.orig_image = scale_im;
      EIG_imgparams = eig_search(EIG_sparams, &X_minima,
				 &Y_minima, &ERR_minima, &N_minima,
				 eig_variances, eig_umap_musigma,
				 eig_spatial_musigma, w, &num_minima);
    }

  /* given targets, picks out features.  Like with eigsearch, no need */
  /* to repeat ml_detect if doing hierarchical head detection. */
  
  MLD_sparams.orig_image = scale_im;
  if (!hier_flag) {

timestamp("pipe:ml_detect:1 (not hierarchical)");

    MLD_imgparams = ml_detect(MLD_sparams, X_minima, Y_minima, ERR_minima,
			      N_minima);
  }

  /* aligns so all features are in same place */

timestamp("pipe:model_warp:1");

  model_warp(MLD_imgparams, MLD_sparams, image_out);
  
  /* applies mask */

timestamp("pipe:mask_image:1");

  mask_image(image_out, mask);
  
  /* applies contrast normalization */

  MS_sparams.orig_image = image_out;
  contrast_map(MS_sparams, image_out);

  if (!hier_flag)
    {
      free_ivector(N_minima, 1, 4);
      free_imatrix(X_minima, 1, 4, 1, num_minima);
      free_imatrix(Y_minima, 1, 4, 1, num_minima);
      free_matrix(ERR_minima, 1, 4, 1, num_minima);
    }
  free_matrix(scale_im,1,SCALEWARP_NROW,1,SCALEWARP_NCOL);
  return;

} /* pipe */

/*------------------------------------------------------------------------*/

void mask_image(image_t image, float **mask)
{
  int     i, j;

  /* multiply image_in by mask */

  for (i = 1; i <= SCALEWARP_NROW; i++)
    for (j = 1; j <= SCALEWARP_NCOL; j++)
      image[i][j] = image[i][j] * mask[i][j];

  return;

} /* mask_image */

/*------------------------------------------------------------------------*/

/***
  Locates a face in the image passed via MS_sparams including, if 
  hier_flag==1, "hierarchical head detection" (i.e. check head has features 
  appropriately placed). 

  n.b. If hier_flag!=1, processed_image will contain the appropriately
  windowed out image patch containing the head. Otherwise it will be left
  unchanged by this function.

  MAS
***/

imagparams_t 
pipe_locate(srchparams_t MS_sparams, 
	    float **ms_variances, float **ms_umap_musigma[MAX_NUM_TEMPLATES],
	    float **ms_spatial_musigma[MAX_NUM_TEMPLATES],
	    srchparams_t EIG_sparams, float **eig_variances, 
	    float **eig_umap_musigma[MAX_NUM_TEMPLATES], 
	    float **eig_spatial_musigma[MAX_NUM_TEMPLATES], float **w,
	    srchparams_t MLD_sparams, image_t image_out,
	    char filename[MAX_CHARS], float **mask, int de_flag, int hier_flag)
{
  imagparams_t MS_imgparams, EIG_imgparams, MLD_imgparams, ret_imgparams;
  float        **ERR_minima;
  int          **X_minima, **Y_minima;
  int          *N_minima;
  char         head_file[MAX_CHARS];
  int          num_minima;
  image_t      scale_im;
  float        **W, **Winv; 
  float        scale;
  int          i,j;

  timestamp("pipe_locate:ms_search:1");

  /* head finding */

  MS_imgparams = ms_search(MS_sparams, EIG_sparams, MLD_sparams, 
			   &MLD_imgparams, hier_flag, ms_variances, 
			   ms_umap_musigma, ms_spatial_musigma, eig_variances,
			   eig_umap_musigma, eig_spatial_musigma, w);

  /* feature looking (picks out possible targets.  If doing */
  /* hierarchical head detection, you already have have the locations */
  /* for the features from that, so no need to do it again. */

  if (!hier_flag) {

      timestamp("pipe_locate:eig_search:1 (not hierarchical)");

      scale_im = matrix(1,SCALEWARP_NROW,1,SCALEWARP_NCOL);
      scale_warp(MS_sparams, MS_imgparams.best_head_x, MS_imgparams.best_head_y,
		 MS_imgparams.best_scale, scale_im);
      
      EIG_sparams.orig_image = scale_im;
      EIG_imgparams = eig_search(EIG_sparams, &X_minima,
				 &Y_minima, &ERR_minima, &N_minima,
				 eig_variances, eig_umap_musigma,
				 eig_spatial_musigma, w, &num_minima);

      /* given targets, picks out features.  Like with eigsearch, no need */
      /* to repeat ml_detect if doing hierarchical head detection. */

      /* --- DEBUG for ml_detect --- */
      { int i;

	write_float_as_uchar(scale_im, 128, 128, "./headimage.raw");
	for (i=1; i<=4; i++)
	  fprintf(stderr,"N_minima[%d] = %d \n", i, N_minima[i]);
	
	fprintf(stderr,"EIG_sparams.do_spatial = %d \n", 
		EIG_sparams.do_spatial);
	fprintf(stderr,"EIG_sparams.do_correlation = %d \n", 
		EIG_sparams.do_correlation);
	fflush(stderr);
      }

      timestamp("pipe_locate:ml_detect:1 (not hierarchical)");

      MLD_sparams.orig_image = scale_im;
      MLD_imgparams = ml_detect(MLD_sparams, X_minima, Y_minima, 
				ERR_minima, N_minima);

      free_ivector(N_minima, 1, 4);
      free_imatrix(X_minima, 1, 4, 1, num_minima);
      free_imatrix(Y_minima, 1, 4, 1, num_minima);
      free_matrix(ERR_minima, 1, 4, 1, num_minima);

      free_matrix(scale_im,1,SCALEWARP_NROW,1,SCALEWARP_NCOL);

  } /* if !hier_flag */

  /* head location details */
  ret_imgparams.best_head_x  = MS_imgparams.best_head_x;
  ret_imgparams.best_head_y  = MS_imgparams.best_head_y;
  ret_imgparams.best_scale   = MS_imgparams.best_scale;

  /* need to map feature positions in "head scalewarped" image back */
  /* into the original image coordinate frame                       */

  /* compute forward mapping matrix */
  W    = matrix(1, 3, 1, 3);
  Winv = matrix(1, 3, 1, 3);
      
  for (i=1; i<=3; i++) 
    for (j=1; j<=3; j++)
      W[i][j] = 0.0;

  scale = MS_imgparams.best_scale * SCALEFACTOR;
  W[1][1] = W[2][2] = scale;
  W[3][3] = 1.0;
  W[1][3] = X_DEST - scale * MS_imgparams.best_head_x;
  W[2][3] = Y_DEST - scale * MS_imgparams.best_head_y;

  /* compute reverse mapping matrix */
  matrix_inverse(W, 3, Winv);

  /* map each feature point into original image coordinates */

  i  = MLD_imgparams.lefteye_x;
  j  = MLD_imgparams.lefteye_y;
  ret_imgparams.lefteye_x = Winv[1][1]*i + Winv[1][2]*j + Winv[1][3];
  ret_imgparams.lefteye_y = Winv[2][1]*i + Winv[2][2]*j + Winv[2][3];

  i = MLD_imgparams.righteye_x;
  j = MLD_imgparams.righteye_y;
  ret_imgparams.righteye_x = Winv[1][1]*i + Winv[1][2]*j + Winv[1][3];
  ret_imgparams.righteye_y = Winv[2][1]*i + Winv[2][2]*j + Winv[2][3];

  i = MLD_imgparams.nose_x;
  j = MLD_imgparams.nose_y;
  ret_imgparams.nose_x = Winv[1][1]*i + Winv[1][2]*j + Winv[1][3];
  ret_imgparams.nose_y = Winv[2][1]*i + Winv[2][2]*j + Winv[2][3];

  i = MLD_imgparams.mouth_x;
  j = MLD_imgparams.mouth_y;
  ret_imgparams.mouth_x = Winv[1][1]*i + Winv[1][2]*j + Winv[1][3];
  ret_imgparams.mouth_y = Winv[2][1]*i + Winv[2][2]*j + Winv[2][3];

  ret_imgparams.energy      = MLD_imgparams.energy;

  free_matrix(W, 1, 3, 1, 3);
  free_matrix(Winv, 1, 3, 1, 3);

  return( ret_imgparams );
 
} /* pipe_locate */

/*------------------------------------------------------------------------*/

/***
  MAS
***/
void
normalise_face(srchparams_t MLD_sparams, imagparams_t MLD_imgparams, 
	       float **mask, image_t image_out )
{

    timestamp("normalise_face:model_warp:1");
  
    /* aligns so all features are in same place in standard sized image */
    model_warp_2(MLD_imgparams, MLD_sparams, image_out, 
		 SCALEWARP_NROW, SCALEWARP_NCOL);

    timestamp("normalise_face:mask_image:1");

    /* applies mask */
    mask_image(image_out, mask);
  
    /* applies contrast normalization */
    MLD_sparams.orig_image = image_out;
    contrast_map(MLD_sparams, image_out);
    
    write_float_as_uchar(image_out, 128, 128, "./normimage.raw");
   
    return;

} /* normalise_face */

/*------------------------------------------------------------------------*/

