#include "util.h"
#include <math.h>
#define PI 3.1415926535

/* make_G0 returns a square matrix of a G0 filter */

int make_G0(int lo, int hi, float **g0, float sigma)
{
  int i,j;
  float *x,scale,s2,norm=0;

  x = vector(lo,hi);
  s2 = sigma*sigma;
  scale = 1.0/(2.0*PI*s2);
 
  i = (int) (hi-lo+1)/2.0; 
  x[lo] = (-1.0 * i);
  for(i=lo+1;i<=hi;i++)
    x[i] = x[i-1] + 1.0;
  
  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) 
      g0[i][j] = scale * exp((-x[i]*x[i]-x[j]*x[j])/(2.0*s2));

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++)
     norm += g0[i][j]*g0[i][j];
  norm = sqrt(norm);
  
  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++)
      g0[i][j] /= norm;
  
  free_vector(x,lo,hi);
  
  return 0;
}

/* make_G1 returns 2 square matrices for a steerable G1 filter */

int make_G1(int lo, int hi, float **g1a, float **g1b, float sigma)
{
  int i,j;
  float *x,**g0,scale,s2,norm=0;

  x = vector(lo,hi);
  g0 = matrix(lo,hi,lo,hi);
  s2 = sigma*sigma;
  scale = 1.0/(2.0*PI*s2);

  i = (int) (hi-lo+1)/2.0;
  x[lo] = (-1.0 * i);
  for(i=lo+1;i<=hi;i++)
    x[i] = x[i-1] + 1.0;

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g0[i][j] = scale * exp((-x[i]*x[i]-x[j]*x[j])/(2.0*s2));
      g1a[i][j] = (-1*x[i]/s2) * g0[i][j];
      g1b[j][i] = g1a[i][j];
    }

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++)
     norm += g1a[i][j]*g1a[i][j];
  norm = sqrt(norm);

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g1a[i][j] /= norm;
      g1b[i][j] /= norm;
    }

  free_vector(x,lo,hi);
  free_matrix(g0,lo,hi,lo,hi);

  return 0;
}

/* make_G2 returns 3 square matrices for a steerable G2 filter */

int make_G2(int lo, int hi, float **g2a, float **g2b, float **g2c, float sigma)
{
  int i,j;
  float *x,**g0,scale,s2,s4,norm=0;

  x = vector(lo,hi);
  g0 = matrix(lo,hi,lo,hi);
  s2 = sigma*sigma;
  s4 = s2*s2;
  scale = 1.0/(2.0*PI*s2);

  i = (int) (hi-lo+1)/2.0;
  x[lo] = (-1.0 * i);
  for(i=lo+1;i<=hi;i++)
    x[i] = x[i-1] + 1.0;

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g0[i][j] = scale * exp((-x[i]*x[i]-x[j]*x[j])/(2.0*s2));
      g2a[i][j] = (x[i]*x[i]/s4 - 1.0/s2) * g0[i][j];
      g2c[j][i] = g2a[i][j];
      g2b[i][j] = (x[i]*x[j]/s4) * g0[i][j];
    }

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++)
     norm += g2a[i][j]*g2a[i][j];
  norm = sqrt(norm);

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g2a[i][j] /= norm;
      g2b[i][j] /= norm;
      g2c[i][j] /= norm;
    }

  free_vector(x,lo,hi);
  free_matrix(g0,lo,hi,lo,hi);

  return 0;
}

/* make_G3 returns 4 square matrices for a steerable G3 filter */

int make_G3(int lo, int hi, float **g3a, float **g3b, float **g3c,
            float **g3d, float sigma)
{
  int i,j;
  float *x,**g0,scale,s2,s4,s6,norm=0;

  x = vector(lo,hi);
  g0 = matrix(lo,hi,lo,hi);
  s2 = sigma*sigma;
  s4 = s2*s2;
  s6 = s4*s2;
  scale = 1.0/(2.0*PI*s2);

  i = (int) (hi-lo+1)/2.0;
  x[lo] = (-1.0 * i);
  for(i=lo+1;i<=hi;i++)
    x[i] = x[i-1] + 1.0;

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g0[i][j] = scale * exp((-x[i]*x[i]-x[j]*x[j])/(2.0*s2));
      g3a[i][j] = (-1.0*x[i]*x[i]*x[i]/s6 + 3.0*x[i]/s4) * g0[i][j];
      g3d[j][i] = g3a[i][j];
      g3b[i][j] = (-1.0*x[i]*x[i]/s4+1.0/s2) * (x[j]/s2) * g0[i][j];
      g3c[j][i] = g3b[i][j];
    }

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++)
     norm += g3a[i][j]*g3a[i][j];
  norm = sqrt(norm);

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g3a[i][j] /= norm;
      g3b[i][j] /= norm;
      g3c[i][j] /= norm;
      g3d[i][j] /= norm;
    }

  free_vector(x,lo,hi);
  free_matrix(g0,lo,hi,lo,hi);

  return 0;
}

/* make_G4 returns the 5 square basis matrices for a steerable G4 filter */

int make_G4(int lo, int hi, float **g4a, float **g4b, float **g4c,
            float **g4d, float **g4e, float sigma)
{
  int i,j;
  float *x,**g0,scale,s2,s4,s6,s8,norm=0;

  x = vector(lo,hi);
  g0 = matrix(lo,hi,lo,hi);
  s2 = sigma*sigma;
  s4 = s2*s2;
  s6 = s4*s2;
  s8 = s4*s4;
  scale = 1.0/(2.0*PI*s2);

  i = (int) (hi-lo+1)/2.0;
  x[lo] = (-1.0 * i);
  for(i=lo+1;i<=hi;i++)
    x[i] = x[i-1] + 1.0;

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g0[i][j] = scale * exp((-x[i]*x[i]-x[j]*x[j])/(2*s2));
      g4a[i][j] = (x[i]*x[i]*x[i]*x[i]/s8-6*x[i]*x[i]/s6+3/s4) * g0[i][j];
      g4e[j][i] = g4a[i][j];
      g4b[i][j] = (-1*x[j]/s2)*(-1*x[i]*x[i]*x[i]/s6+3*x[i]/s4)* g0[i][j];
      g4d[j][i] = g4b[i][j];
      g4c[i][j] = (x[i]*x[i]/s4-1/s2) * (x[j]*x[j]/s4-1/s2) * g0[i][j];
    }

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++)
     norm += g4a[i][j]*g4a[i][j];
  norm = sqrt(norm);

  for(i=lo;i<=hi;i++)
    for(j=lo;j<=hi;j++) {
      g4a[i][j] /= norm;
      g4b[i][j] /= norm;
      g4c[i][j] /= norm;
      g4d[i][j] /= norm;
      g4e[i][j] /= norm;
    }

  free_vector(x,lo,hi);
  free_matrix(g0,lo,hi,lo,hi);

  return 0;
}

