/* ---------------------------------------------------------------------- 
   PROGRAM: io.c
   DATE:    10/24/93
   AUTHOR:  Baback Moghaddam, baback@media.mit.edu

   File I/O Routines
------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/stat.h>

#include "util.h"

#define MAX_CHARS   256   /* max number of chars per line */

/* machine-sizes for basic types  */

#define  CHAR_SIZE		sizeof(char)
#define  SHORT_INT_SIZE		sizeof(short int)
#define  INT_SIZE  		sizeof(int)
#define  LONG_INT_SIZE		sizeof(long int)
#define  FLOAT_SIZE		sizeof(float)
#define  DOUBLE_SIZE		sizeof(double)


/*----------------------------------------------------------------*/
#define FLOATBYTESWAP(x, y) \
{   ((char *) (y))[3] = ((char *) (x))[0];\
    ((char *) (y))[0] = ((char *) (x))[3];\
    ((char *) (y))[2] = ((char *) (x))[1];\
    ((char *) (y))[1] = ((char *) (x))[2];    }
/*----------------------------------------------------------------*/


/* -------------------------------------------------------- */
/* Routines for reading and writing PGM (Raw) format images */
/* -------------------------------------------------------- */

float **read_PGM(char *filename, int *height, int *width)
{
  FILE *fp;
  register int i,j;
  int n;
  char str[10];
  unsigned char *row;
  float **image;

  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"ERROR: read_PGM could not open file %s\n\n",filename);
    exit(1);
  }
  fscanf(fp,"%s",str);
  if (strcmp(str,"P5")) {
    fprintf(stderr,"ERROR: Routine read_PGM found non-PGMRaw file.");
    exit(1);
  }
  fscanf(fp,"%d %d",width,height);
  fscanf(fp,"%d",&n);
  fgetc(fp);  /* skip over carriage return */

  image = matrix(1, *height, 1, *width);
  row = cvector(1, *width);
  
  /* NOTE: ----------- have to fix the row+1 to row+offset to allow for
                       variable offset arrays (like 0-offset as opposed
                       to unit-offset all the time!).                   */

  for (i=1;i<=(*height);i++) {
   fread(row+1, CHAR_SIZE, *width, fp);  
   for (j=1;j<=(*width);j++)
    image[i][j] = (float) row[j];
  }

  free_cvector(row, 1, *width);
  fclose(fp);
  return image;
}

int write_PGM(char *filename, unsigned char **image, int height, int width)
{
  FILE *fp;
  int i,j,n;

  if ( (fp = fopen(filename, "w")) == NULL) {
    fprintf(stderr,"ERROR: write_PGM could not open file %s\n\n", filename);
    exit(1);
  }
  fprintf(fp,"P5\n",filename);
  fprintf(fp,"%d %d\n",width,height);
  fprintf(fp,"%d\n",255);

  /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
		       allow for variable offset arrays (like 0-offset
		       as opposed to unit-offset all the time!).          */

  for (i=1;i<=height;i++) 
    fwrite(image[i]+1, CHAR_SIZE, width, fp);

 fclose(fp);
 return 0;
}


/* -------------------------------------------------------- */
/*     Routines for reading and writing RAW format images   */
/* -------------------------------------------------------- */
/*  NOTE: These routines take a pre-allocated float **image */
/* -------------------------------------------------------- */

int read_RAW(float **image, char *filename, int height, int width)
{
  FILE *fp;
  register int i,j;
  int n;
  unsigned char *row;
  
  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"ERROR: read_RAW could not open file %s\n\n",filename);
    exit(1);
  }
  
  row = cvector(1, width);
  
  /* NOTE: ----------- have to fix the row+1 to row+offset to allow for
                       variable offset arrays (like 0-offset as opposed
                       to unit-offset all the time!).                   */

  for (i=1;i<=height;i++) {
   fread(row+1, CHAR_SIZE, width, fp);
   for (j=1;j<=width;j++)
    image[i][j] = (float) row[j];
  }

  free_cvector(row, 1, width);
  fclose(fp);
  return 0;
}


int read_RAW_float(float **image, char *filename, int height, int width)
{
  FILE *fp;
  register int i,j;
  int n;
  float f0;

  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"ERROR: read_RAW_float could not open file %s\n\n",
	    filename);
    exit(1);
  }

  /* NOTE: ----------- have to fix the row+1 to row+offset to allow for
                       variable offset arrays (like 0-offset as opposed
                       to unit-offset all the time!).                   */

  
  for (i=1;i<=height;i++) 
    fread(image[i]+1, FLOAT_SIZE, width, fp); 
  fclose(fp);  


  if (strcmp(getenv("FLOAT_SWAP"),"on") == 0) {
    
    for (i=1;i<=height;i++) {
      for (j=1; j<=width; j++) {
	FLOATBYTESWAP(image[i]+j, &f0);
	image[i][j] = f0;
      }
    }
  }
  
  return 0;
}

int write_RAW(char *filename, unsigned char **image, int height, int width)
{
  FILE *fp;
  int i,j,n;

  if ( (fp = fopen(filename, "w")) == NULL) {
    fprintf(stderr,"ERROR: write_RAW could not open file %s\n\n",filename);
    exit(1);
  }

  /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
                       allow for variable offset arrays (like 0-offset
                       as opposed to unit-offset all the time!).          */

    for (i=1;i<=height;i++)
      fwrite(image[i]+1, CHAR_SIZE, width, fp); 

 fclose(fp);
 return 0;
}


int write_RAW_float(char *filename, float **image, int height, int width)
{
  FILE *fp;
  int i,j,n;
  float *row;

  if ( (fp = fopen(filename, "w")) == NULL) {
    fprintf(stderr,"ERROR: write_RAW_float could not open file %s\n\n",
	    filename);
    exit(1);
  }

  /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
                       allow for variable offset arrays (like 0-offset
                       as opposed to unit-offset all the time!).          */

  if (strcmp(getenv("FLOAT_SWAP"),"on") != 0) {

    for (i=1;i<=height;i++)
      fwrite(image[i]+1, FLOAT_SIZE, width, fp); }

  else {

    row = vector(1, width);
    for (i=1;i<=height;i++) {
      for (j=1; j<=width; j++) 
	FLOATBYTESWAP(image[i]+j, row+j);
      fwrite(row+1, FLOAT_SIZE, width, fp); 
    }
    free_vector(row, 1, width);
  }
  
 fclose(fp);
 return 0;
}


/* -------------------------------------------------------- */
/* Routines for reading and writing Binary format files     */
/* -------------------------------------------------------- */

float **read_BIN(char *filename, int *height, int *width)
{
  FILE *fp;
  register int i,j;
  float **image, temp[2],f0,f1;

  if ( (fp = fopen(filename,"r")) == NULL) {
    fprintf(stderr,"ERROR: read_BIN could not open file %s\n\n", filename);
    exit(1);
  }

  if (strcmp(getenv("FLOAT_SWAP"),"on") != 0) {     
    
    fread(temp, FLOAT_SIZE, 2, fp);
    *height = (int) temp[0];
    *width  = (int) temp[1];
    
    image = matrix(1, *height, 1, *width);
    
    /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
                         allow for variable offset arrays (like 0-offset
                         as opposed to unit-offset all the time!).          */
    
    for (i=1;i<=(*height);i++) 
      fread(image[i]+1, FLOAT_SIZE, *width, fp); }
  
  else {   /* SWAP Endian-Order */
   
      fread(temp, FLOAT_SIZE, 2, fp);
      FLOATBYTESWAP(temp+0, &f0);
      FLOATBYTESWAP(temp+1, &f1);
      *height = (int) f0;
      *width  = (int) f1;
      
      image = matrix(1, *height, 1, *width);
      
      /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
	                   allow for variable offset arrays (like 0-offset
	                   as opposed to unit-offset all the time!).       */
      
      for (i=1;i<=(*height);i++) { 
	fread(image[i]+1, FLOAT_SIZE, *width, fp);
	for (j=1; j<=(*width); j++) {
	  FLOATBYTESWAP(image[i]+j, &f0);
	  image[i][j] = f0;
	}
      }

    }
  
  fclose(fp);
  return image;
}

int write_BIN(char *filename, float **image, int height, int width)
{
  FILE *fp;
  register int i,j;
  float temp[2], f0, f1, *row; 

  if ((fp = fopen(filename,"w")) == NULL) {
    fprintf(stderr,"ERROR: write_BIN could not open file %s\n\n",filename);
    exit(1);
  }

  if (strcmp(getenv("FLOAT_SWAP"),"on") != 0) {     
    
    temp[0] = height;
    temp[1] = width;
    fwrite(temp, FLOAT_SIZE, 2, fp);
    
    /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
                         allow for variable offset arrays (like 0-offset
                         as opposed to unit-offset all the time!).          */

    for (i=1;i<=height;i++) 
      fwrite(image[i]+1, FLOAT_SIZE, width, fp); }
  
  else {  /* swap Endian-Order */

    f0 = height;
    f1 = width;
    FLOATBYTESWAP(&f0, temp+0);
    FLOATBYTESWAP(&f1, temp+1);
    fwrite(temp, FLOAT_SIZE, 2, fp);
    
    /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
                         allow for variable offset arrays (like 0-offset
                         as opposed to unit-offset all the time!).          */


    row = vector(1, width);
    for (i=1;i<=height;i++) {
      for (j=1; j<=width; j++) 
	FLOATBYTESWAP(image[i]+j, row+j);
      fwrite(row+1, FLOAT_SIZE, width, fp); 
    }

    free_vector(row, 1, width);
  }
  
  
  fclose(fp); 
  return 0;
}


/*-----------------------------------------------------------*/
/*         Routines for reading/writing DAT files            */
/*-----------------------------------------------------------*/

void read_descriptor(char *dirname, int *nframe, int *sets,
		     int *bytes_pixel, int *xdim, int *ydim)
{
  char label[20], junk[20], junk2[20], line[80];
  char *filename;
  FILE *ifs_desc;

  filename = (char *) malloc(strlen(dirname) + 13);
  sprintf(filename,"%s/descriptor", dirname);
  
  if ((ifs_desc = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"\nCannot open input descriptor file: %s\n", filename);
    exit(1);
  }

  *nframe = 1;      /* default to 1 */
  *bytes_pixel = 1; /* default to 1 */
  *sets = 1;        /* default to 1 */

  while (fgets(line,80,ifs_desc) != NULL) {
    
    sscanf(line,"%s",label);
    if (!strcmp (label, "(_data_files")) 
      sscanf(line,"%s%d%s", junk, nframe, junk2);
    else  if (!strcmp (label, "(_dimensions"))
      sscanf(line,"%s%d%d%s", junk, ydim, xdim, junk2);
    else  if (!strcmp (label, "(_data_sets"))
      sscanf(line,"%s%d%s", junk, sets, junk2);
    else  if (!strcmp (label, "(_data_type")) {
	sscanf(line,"%s%s", junk, junk2);
	if ((!strncmp (junk2, "\"ascii\")", 5)) ||
	    (!strncmp (junk2, "ascii", 5)))
	  *bytes_pixel = 0;
	if ((!strncmp (junk2, "\"unsigned_1\")", 5)) ||
	    (!strncmp (junk2, "unsigned_1", 5)))
	  *bytes_pixel = 1;
	if ((!strncmp (junk2, "\"float\")", 5)) ||
	    (!strncmp (junk2, "float", 5)))
	  *bytes_pixel = 4;
      }

  }

  fclose (ifs_desc);
  free(filename);
}

/* ---------------------------------------------------------------- */

void write_descriptor(char *dirname, int nframe, 
		      int xdim, int ydim,
		      int bytes_pixel, char *history)

{
  char *dat_type;
  char *filename;
  FILE *ofs_desc;
  
  filename = (char *) malloc(strlen(dirname) + 13);
  sprintf(filename,"%s/descriptor", dirname);

  if ((ofs_desc = fopen(filename, "w")) == NULL) {
    fprintf(stderr,"\nCannot open output descriptor file: %s\n", filename);
    exit(1);
  }
  
  dat_type = (char *) malloc(sizeof(char) * 30);
  
  if (bytes_pixel == 4)
    strcpy(dat_type, "\"float\"");
  if (bytes_pixel == 1)
    strcpy(dat_type, "\"unsigned_1\"");
  if (bytes_pixel == 0)
    strcpy(dat_type, "\"ascii\"");

  fprintf(ofs_desc,"(_data_files %d)\n(_channels 1)\n(_dimensions %d %d)\n(_data_type %s)\n(_history   \"%s\")\n",nframe, ydim, xdim, dat_type, history);
  
  fclose(ofs_desc);
  free(filename);
}

/* -------------------------------------------------------- */

void read_DAT_frame(float **image, char *dirname,
		    int frame, int nframe, int height, int width)
     
{
  FILE *fp;
  register int i,j;
  int n;
  char *filename;
  unsigned char *row;

  filename = (char *) malloc(strlen(dirname) + 8);
  if (nframe>1) 
    sprintf(filename,"%s/data%d", dirname, frame);
  else
    sprintf(filename,"%s/data", dirname);

  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"\nCould not open input file: %s ... Exiting\n\n",filename);
    exit(1);
  }
  
  row = cvector(1, width);
  
  /* NOTE: ----------- have to fix the row+1 to row+offset to allow for
                       variable offset arrays (like 0-offset as opposed
                       to unit-offset all the time!).                   */
  
  for (i=1;i<=height;i++) {
    fread(row+1, CHAR_SIZE, width, fp);
    for (j=1;j<=width;j++)
      image[i][j] = (float) row[j];
  }
  
  free_cvector(row, 1, width);
  
  fclose(fp);
  free(filename);
  
}
  
void write_DAT_frame(unsigned char **image, char *dirname, 
		     int frame, int nframe, int height, int width)
{
  FILE *fp;
  int i,j,n;
  char *filename;

  filename = (char *) malloc(strlen(dirname) + 8);
  if (nframe>1) 
    sprintf(filename,"%s/data%d", dirname, frame);
  else
    sprintf(filename,"%s/data", dirname);

  if ((fp = fopen(filename, "w")) == NULL) {
    fprintf(stderr,"\nCould not open output file: %s ... Exiting\n ",filename);
    exit(1);
  }

  /* NOTE: ----------- have to fix the image[i]+1 to image[i]+offset to
                       allow for variable offset arrays (like 0-offset
                       as opposed to unit-offset all the time!).          */

  for (i=1;i<=height;i++)
    fwrite(image[i]+1, CHAR_SIZE, width, fp);

 fclose(fp);
 free(filename);

}



/* ====================================================================

   Routines for reading and writing ASCII-format (matlab ascii) matrices
  
   ==================================================================== */

int get_ASCII_dimensions(char *filename, int *nrow, int *ncol)

/* reads (and checks) the dimensions of an ASCII matrix (nrow by ncol)
   it returns 0 if all nrow lines have ncol tokens, 1 if not           */
 
{

  int i,j,M,N;
  FILE *fp;
  char line[MAX_CHARS], str[MAX_CHARS];


  /* ---- open file and count # of lines ------- */

  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"ERROR: get_ASCII_dimensions() could not open %s\n\n", filename);
    exit(1);
  }
  M = 0;
  while (strlen(fgets(line, MAX_CHARS, fp))) 
    M++;
  rewind(fp);
  *nrow = M;


  /* ---- count # of columns in a line ---- */
  
  fgets(line, MAX_CHARS, fp);
  strtok(line, " ");
  N = 1;
  while (strtok(NULL, " ") != NULL)
    N++;
  rewind(fp);
  *ncol = N;


  /* --- check that all lines have N tokens --- */

  for (i=1; i<=M; i++) {
    fgets(line, MAX_CHARS, fp);
    strtok(line, " ");
    j = 1;
    while (strtok(NULL, " ") != NULL)
      j++;
    if (j!=N) {
      fclose(fp);
      return 1;
    } 
  }

  fclose(fp);
  return 0;
}

   

  

int read_ASCII(char *filename, float **image, int nrow, int ncol)

/*
  reads an ASCII matrix (nrow by ncol) into the (pre-allocated)
  matrix "image" from the file "filename".
  
  If insufficient lines or tokens are encountered it returns 1
  else it returns 0.

*/
 
{

  register int i,j;
  FILE *fp;
  char line[MAX_CHARS], str[MAX_CHARS];


  /* ---- open file  ------- */

  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"ERROR: read_ASCII() could not open %s\n\n", filename);
    exit(1);
  }
  
  i = 0;
  while (strlen(fgets(line, MAX_CHARS, fp))) {
    i++;
    image[i][1] = atof(strtok(line, " "));
    for (j=2; j<=ncol; j++) {
      strcpy(str, strtok(NULL, " "));
      image[i][j] = atof(str);

    }
  }
  

  fclose(fp);
  
  if (i!=nrow)
    return 1;
  else
    return 0;
}

  

int write_ASCII(char *filename, float **image, int nrow, int ncol)

/*
  writes an ASCII matrix (nrow by ncol) to disk as filename

*/
 
{

  register int i,j;
  FILE *fp;
  char line[MAX_CHARS], str[MAX_CHARS];


  /* ---- open file  ------- */

  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr,"ERROR: read_ASCII() could not open %s\n\n", filename);
    exit(1);
  }


  for (i=1; i<=nrow; i++) {
    for (j=1; j<=ncol; j++)
      fprintf(fp, "%f ", image[i][j]);
    fprintf(fp, "\n");
  }

  fclose(fp);
  return 0;
}

