/*----------------------------------------------------------------------
PROGRAM: matrix.c
DATE:    3/14/94
AUTHOR:  Baback Moghaddam, baback@media.mit.edu
------------------------------------------------------------------------

  Matrix routines

  NOTE: uses gaussj() and svdcmp() from libnrc.a 
---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "util.h"




/*----------------------------------------------------------------------*/

void matrix_multiply(float **A, float **B, int M, int N, int P,
		     float **C)
     
     /*  matrix multiply of A[1..M][1..N] and B[1..N][1..P] 
	 the result C[1..M][1..P] must be pre-allocated
	 */

{
  register int i,j,k;
  float sum;

  for (i=1; i<=M; i++)
    for (j=1; j<=P; j++) {
      for (k=1, sum=0.0; k<=N; k++)
	sum += A[i][k]*B[k][j];
      C[i][j] = sum;
    }
}


/*----------------------------------------------------------------------*/

void matrix_inverse(float **A, int N, float **Ainv)

     /* matrix inversion routine using gaussj()  */

{

  register int i,j;
  float **myA, **b;

  myA = matrix(1, N, 1, N);
  b   = matrix(1, N, 1, 1);

  for (i=1; i<=N; i++)
    for (j=1; j<=N; j++) {
      myA[i][j] = A[i][j];
      b[j][1] = j;
    }

  gaussj(myA, N, b, 1);

  for (i=1; i<=N; i++)
    for (j=1; j<=N; j++) 
      Ainv[i][j] = myA[i][j];

  free_matrix(myA, 1, N, 1, N);
  free_matrix(b,   1, N, 1, 1);

}


/*----------------------------------------------------------------------*/

float matrix_norm(float **A, int M, int N)

     /* returns the matrix norm (largest singular value) of A */

{

  register int i,j;
  float **myA, **V, *s, smax;
  int m, n;

  if (M>N) {
    m = M;
    n = N;
    myA = matrix(1, m, 1, n);
    for (i=1; i<=m; i++)
      for (j=1; j<=n; j++)
	myA[i][j] = A[i][j];
  }
  else {
    m = N;
    n = M;
    myA = matrix(1, m, 1, n);
    for (i=1; i<=m; i++)
      for (j=1; j<=n; j++)
	myA[i][j] = A[j][i];
  }

  s = vector(1, n);
  V = matrix(1, n, 1, n);
  svdcmp(myA, m, n, s, V);
  smax = -FLT_MIN;
  for (i=1; i<=n; i++)
    if (s[i]>smax) smax=s[i];

  free_vector(s, 1, n);
  free_matrix(V, 1, n, 1, n);
  free_matrix(myA, 1, m, 1, n);

  return smax;
}


/*----------------------------------------------------------------------*/

void matrix_element_multiply(float **A, float **B, 
			     int M, int N, float **C)

     /*  element-by-element multiplication C = A .* B
	 the result C must be pre-allocated
	 */

{
  register int i,j;
  
  for (i=1; i<=M; i++)
    for (j=1; j<=N; j++)
      C[i][j] = A[i][j] * B[i][j];
}

/*----------------------------------------------------------------------*/

float matrix_element_sum(float **A, int M, int N)

     /*  summation of the elements of matrix A */

{
  register int i,j;
  float sum = 0.0;

  for (i=1; i<=M; i++)
    for (j=1; j<=N; j++)
      sum += A[i][j];

  return sum;
}


/* ----------------------------------------------------------------------*/

float matrix_element_max(float **A, 
			 int m1, int m2,
			 int n1, int n2)

     /*  returns value of the maximum element of
	 A[m1...m2][n1...n2]

	 */

{
  register int i,j;
  float a = -FLT_MIN;

  for (i=m1; i<=m2; i++)
    for (j=n2; j<=n2; j++)
      if (A[i][j]>a) a=A[i][j];

  return a;
}

/* ----------------------------------------------------------------------*/

float matrix_element_min(float **A,
			 int m1, int m2,
			 int n1, int n2)

     /*  returns value of the minimum element of 
	 A[m1...m2][n1...n2]
	
	 */


{
  register int i,j;
  float a = FLT_MAX;

  for (i=m1; i<=m2; i++)
    for (j=n1; j<=n2; j++)
      if (A[i][j]<a) a=A[i][j];

  return a;
}

/* ----------------------------------------------------------------------*/


float vector_element_max(float *A, int m1, int m2)

     /*  returns value of the maximum element of A[m1...m2]

	 */

{
  register int i;
  float a = -FLT_MIN;

  for (i=m1; i<=m2; i++)
      if (A[i]>a) a=A[i];

  return a;
}

/* ----------------------------------------------------------------------*/

float vector_element_min(float *A, int m1, int m2)

     /*  returns value of the minimum element of A[m1...m2] 

	 */
{
  register int i;
  float a = FLT_MAX;

  for (i=m1; i<=m2; i++)
      if (A[i]<a) a=A[i];

  return a;
}


